"""Tools for managing render contexts.

@author: Stephan Wenger
@date: 2012-03-08
"""

class ContextBindingProxy(object):
    _bound_context = None

    def __get__(self, obj, cls=None):
        return self._bound_context

    def __set__(self, obj, context=None):
        if context is None:
            self._bound_context = None
        elif context != self._bound_context:
            self._bound_context = context
            context._bind()

    def __repr__(self):
        return "proxy for context binding"

class ContextManager(object):
    current_context = ContextBindingProxy()
    """The currently active context."""

    def __init__(self):
        self._stack = []

    def push(self, context):
        self._stack.append(self.current_context)
        self.current_context = context

    def pop(self):
        self.current_context = self._stack.pop()

    @staticmethod
    def create_default_context():
        """Create a default offscreen rendering context.

        @todo: This is window system dependent; create an appropriate context
        dynamically.
        @todo: When no context exists, create a raw offscreen context instead of a
        hidden GLUT window so that rendering is possible without an X connection.
        """

        try:
            from glitter.contexts.glx import GLXContext
            return GLXContext()
        except ImportError:
            from glitter.contexts.glut import GlutWindow
            return GlutWindow(shape=(1, 1), hide=True)

context_manager = ContextManager()

__all__ = ["context_manager"]




"""OpenGL context implementations for various platforms.

@todo: Create a raw offscreen GLX context class with context generation and switching and use it as a default context instead of GLUT.
@todo: Qt GL widgets and contexts.

@author: Stephan Wenger
@date: 2012-02-29
"""

from glitter.contexts import glut
from glitter.contexts.context import *
from glitter.contexts.contextmanager import *




"""Generic descriptor classes for per-context state.

@author: Stephan Wenger
@date: 2012-02-29
"""

from weakref import WeakKeyDictionary

import glitter.raw as _gl
from glitter.utils import hints, bool8, int32, int64, float32, Proxy

class BooleanProxy(Proxy):
    def __init__(self, get_args=(), setter=None, set_args=(), shape=()):
        super(BooleanProxy, self).__init__(_gl.glGetBooleanv, get_args, setter, set_args, bool8, shape)

class FloatProxy(Proxy):
    def __init__(self, get_args=(), setter=None, set_args=(), shape=()):
        super(FloatProxy, self).__init__(_gl.glGetFloatv, get_args, setter, set_args, float32, shape)

class IntegerProxy(Proxy):
    def __init__(self, get_args=(), setter=None, set_args=(), shape=()):
        super(IntegerProxy, self).__init__(_gl.glGetIntegerv, get_args, setter, set_args, int32, shape)

class Integer64Proxy(Proxy):
    def __init__(self, get_args=(), setter=None, set_args=(), shape=()):
        super(Integer64Proxy, self).__init__(_gl.glGetInteger64v, get_args, setter, set_args, int64, shape)

class EnableDisableProxy(Proxy):
    def __init__(self, arg):
        super(EnableDisableProxy, self).__init__(name=arg)
        self._arg = arg

    def __get__(self, obj, cls=None):
        with obj:
            return _gl.glIsEnabled(self._arg) == _gl.GL_TRUE

    def __set__(self, obj, value):
        with obj:
            _gl.glEnable(self._arg) if value else _gl.glDisable(self._arg)

class EnumProxy(Proxy):
    def __init__(self, enum, arg, setter=None, set_args=()):
        super(EnumProxy, self).__init__(enum=enum, getter=_gl.glGetIntegerv, get_args=(arg,), setter=setter, set_args=set_args)

class HintProxy(EnumProxy):
    def __init__(self, hint):
        super(HintProxy, self).__init__(hints, hint, _gl.glHint, [hint])

class StringProxy(Proxy):
    def __init__(self, arg, count_attr=None):
        super(StringProxy, self).__init__(name=arg)
        self._arg = arg
        self._count_attr = count_attr

    def __get__(self, obj, cls=None):
        if self._count_attr is None:
            with obj:
                return _gl.string_at(_gl.glGetString(self._arg))
        else:
            with obj:
                _n = _gl.GLint()
                _gl.glGetIntegerv(self._count_attr, _gl.pointer(_n))
                return [_gl.string_at(_gl.glGetStringi(self._arg, i)) for i in range(_n.value)]

    def __set__(self, obj, value):
        raise AttributeError("can't set attribute")

class BindingProxy(Proxy):
    def __init__(self, setter, set_args=()):
        super(BindingProxy, self).__init__(setter=setter, set_args=set_args)
        self._value = WeakKeyDictionary()

    def __get__(self, obj, cls=None):
        return self._value.get(obj, None)

    def __set__(self, obj, value=None):
        with obj:
            old_value = self._value.get(obj, None)
            self._value[obj] = value
            if old_value is not None and old_value != value and hasattr(old_value, "_on_release") and old_value._on_release is not NotImplemented:
                old_value._on_release()
            try:
                self._setter(*([getattr(obj, x) if isinstance(x, str) else x for x in self._set_args] + [0 if value is None else value._id]))
            except:
                self._value[obj] = old_value
                raise
            else:
                if value is not None and value != old_value and hasattr(value, "_on_bind") and value._on_bind is not NotImplemented:
                    value._on_bind()




"""Pipeline class unifying vertex array, shader and framebuffer.

@author: Stephan Wenger
@date: 2012-03-05
"""

from glitter.framebuffers import Framebuffer
from glitter.arrays import VertexArray
from glitter.utils import ItemProxy, InstanceDescriptorMixin, State, StateMixin, add_proxies, PropertyProxy
from glitter.shaders.uniform import BaseUniform # BaseUniform is an implementation detail of the shaders package

class Pipeline(InstanceDescriptorMixin, StateMixin):
    """Convenience class for rendering pipelines.

    C{Pipeline}s contain a vertex array, a shader, and an optional framebuffer.
    Property access and method calls are appropriately redirected to these
    objects. Vertex attributes, shader uniforms, and fragment outputs can be
    set by their name either in the constructor or by accessing the
    corresponding properties on the pipeline.

    Vertex attributes can be set to buffer objects or anything that can be
    converted to an C{ArrayBuffer}.

    Shader uniforms behave as usual.

    Fragment outputs can be set to texture objects.

    @attention: In order to guarantee correct functioning of vertex attribute,
    uniform and fragment output variables, their names must be unique, must not
    begin with an underscore, and must not collide with the names of any vertex
    array or framebuffer methods.

    Usage examples:

    >>> shader = ShaderProgram(...)
    >>> vertices = [...]
    >>> colors = [...]
    >>> elements = [...]
    >>> texture = Texture2D(...)

    >>> pipeline = Pipeline(shader, in_position=vertices, in_color=colors, elements=elements, out_color=texture)
    >>> pipeline.clear()
    >>> pipeline.draw()

    >>> pipeline = Pipeline(shader, use_framebuffer=False)
    >>> with pipeline(in_position=vertices, in_color=colors, elements=elements):
    ...     pipeline.draw()
    >>> with pipeline(in_position=vertices, in_color=colors, elements=elements) as p:
    ...     p.draw()
    >>> pipeline.draw_with(in_position=vertices, in_color=colors, elements=elements)

    >>> pipeline = Pipeline(shader, use_framebuffer=False)
    >>> vertex_array = VertexArray(...)
    >>> with pipeline:
    ...     vertex_array.draw()
    """

    _frozen = False
    """Whether setting of unknown attributes should be interpreted literally or as accessing vertex array and framebuffer properties."""

    def __init__(self, shader, use_framebuffer=True, **kwargs):
        """Create a new C{Pipeline}.

        @param shader: The compiled and linked shader program object to use.
        @type shader: L{ShaderProgram}
        @param use_framebuffer: If C{True}, render to textures instead of the currently bound framebuffer.
        @type use_framebuffer: C{bool}
        @param kwargs: Named arguments are translated to setting of attributes.

        @todo: C{_vao} and C{_fbo} should be created dynamically when (and if)
        attributes and attachments are accessed, respectively;
        C{use_framebuffer} should not be necessary then.
        """

        self._shader = shader
        self._context = self._shader._context
        self._lazy_context_properties = {}
        self._lazy_context_property_stack = []

        # add shader uniforms (attributes are handled by the vertex array)
        for name, proxy in list(self._shader.__dict__.items()):
            if isinstance(proxy, BaseUniform):
                setattr(self, name, PropertyProxy(self._shader, name))

        # create vertex array
        self._vao = VertexArray(context=self._context)
        add_proxies(self, self._vao)
        
        # create framebuffer
        if use_framebuffer:
            self._fbo = Framebuffer(context=self._context)
            add_proxies(self, self._fbo)
        else:
            self._fbo = None

        self._frozen = True
        
        # translate kwargs to setters on self
        for key, value in list(kwargs.items()):
            setattr(self, key, value)

    def _has_input(self, name):
        """Determine whether the shader as an attribute named C{name}."""
        return self._shader.has_attribute_location(name)

    def _add_input(self, name, value=None):
        """Add a proxy for the attribute named C{name}."""
        index = self._shader.get_attribute_location(name)
        with State(self, _frozen=False):
            setattr(self, name, ItemProxy(self._vao, index))
        setattr(self, name, value)

    def _has_output(self, name):
        """Determine whether the shader as a fragment output named C{name}."""
        return self._fbo is not None and self._shader.has_frag_data_location(name)

    def _add_output(self, name, value):
        """Add a proxy for the fragment output named C{name}."""
        index = self._shader.get_frag_data_location(name)
        with State(self, _frozen=False):
            setattr(self, name, ItemProxy(self._fbo, index))
        setattr(self, name, value)

    def __setattr__(self, name, value):
        """Set an attribute.

        When C{name} is a known attribute or starts with an underscore, or when
        C{self} is not in L{_frozen} state, C{__setattr__} works as usual.

        When C{self} is in L{_frozen} state (default), setting of unknown
        attributes will check for vertex attributes, fragment outputs and
        context properties called C{name} and create appropriate proxies or
        raise an error if no such attribute or output exists.

        Context properties set here will be set on binding and reset on
        unbinding the pipeline.
        """

        if hasattr(self, name) or name.startswith("_") or not self._frozen:
            super(Pipeline, self).__setattr__(name, value)
        elif self._has_input(name):
            self._add_input(name, value)
        elif self._has_output(name):
            self._add_output(name, value)
        elif hasattr(self._context, name):
            self._lazy_context_properties[name] = value
        else:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

    def __delattr__(self, name, value):
        """Delete an attribute.

        When C{name} is a known attribute or starts with an underscore, or when
        C{self} is not in L{_frozen} state, C{__delattr__} works as usual.

        Otherwise, if the attribute is a previously set context property, it
        will be removed from the pipeline.
        """

        if hasattr(self, name) or name.startswith("_") or not self._frozen:
            super(Pipeline, self).__delattr__(name, value)
        elif name in self._lazy_context_properties:
            del self._lazy_context_properties[name]
        else:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

    def __enter__(self):
        """Bind framebuffer and shader."""
        
        stack_frame = []
        for key, value in list(self._lazy_context_properties.items()):
            stack_frame.append((key, getattr(self._context, key)))
            setattr(self._context, key, value)
        self._lazy_context_property_stack.append(stack_frame)

        if self._fbo is not None:
            self._fbo.__enter__()
        self._shader.__enter__()
        return self

    def __exit__(self, type, value, traceback):
        """Unbind framebuffer and shader."""
        self._shader.__exit__(type, value, traceback)
        if self._fbo is not None:
            self._fbo.__exit__(type, value, traceback)

        for key, value in self._lazy_context_property_stack.pop():
            setattr(self._context, key, value)

    def draw_with(self, *args, **kwargs):
        """Call L{draw<VertexArray.draw>} on C{self} with attributes from C{kwargs} set.

        Keyword arguments that are known attributes or shader input our output
        variable names will be set as properties on C{self}. Context properties
        are set before drawing and reset afterwards. Any other arguments will
        be passed to L{draw<VertexArray.draw>}.

        @todo: C{kwargs} context properties should override C{_lazy_context_properties}.
        """

        def is_valid_property(name):
            return hasattr(self, name) or self._has_input(name) or self._has_output(name)
        with self(**{key: kwargs.pop(key) for key, value in list(kwargs.items()) if is_valid_property(key)}):
            with State(context=self._context, **{key: kwargs.pop(key) for key, value in list(kwargs.items()) if hasattr(self._context, key)}):
                self.draw(*args, **kwargs)

__all__ = ["Pipeline"]




"""Convenience wrappers without direct OpenGL interaction.

@author: Stephan Wenger
@date: 2012-03-06
"""

from glitter.convenience.pipeline import *
from glitter.convenience.defaultpipeline import *
from glitter.convenience.copying import *
from glitter.convenience.mesh import *
from glitter.convenience.matrices import *




"""Shader program class.

@todo: wrap C{glGetProgramStage}, C{glGetActiveSubroutineName},
C{glGetActiveSubroutineUniform}, C{glGetActiveSubroutineUniformName},
C{glGetSubroutineIndex}, C{glGetSubroutineUniformLocation},
C{glGetTransformFeedbackVarying}, C{glTransformFeedbackVarying},
C{glGetUniformBlockIndex}, C{glGetUniformIndices}, C{glGetActiveUniformBlock},
C{glGetActiveUniformBlockName}, C{glUniformBlockBinding}, and
C{glUniformSubroutine}.

@author: Stephan Wenger
@date: 2012-02-29
"""

from collections import OrderedDict as _odict
import re as _re

import glitter.raw as _gl
from glitter.utils import transform_feedback_buffer_modes, int32, ShaderDatatype, ManagedObject, BindableObject, ShaderLinkError, ShaderValidateError, ListProxy, InstanceDescriptorMixin, Proxy
from glitter.shaders.shader import Shader, VertexShader, TesselationControlShader, TesselationEvaluationShader, GeometryShader, FragmentShader
from glitter.shaders.attribute import Attribute, AttributeStruct, AttributeStructArray
from glitter.shaders.uniform import Uniform, UniformStruct, UniformStructArray

class ProgramProxy(Proxy):
    def __init__(self, _id, arg, enum=None):
        super(ProgramProxy, self).__init__(_gl.glGetProgramiv, [_id, arg], dtype=int32, enum=enum)

class ShaderProgram(ManagedObject, BindableObject, InstanceDescriptorMixin):
    _generate_id = _gl.glCreateProgram
    _delete_id = _gl.glDeleteProgram
    _db = "shader_programs"
    _binding = "current_program"

    transform_feedback_buffer_modes = transform_feedback_buffer_modes
    _frozen = False

    def __init__(self, shaders=[], vertex=[], tess_control=[], tess_evaluation=[], geometry=[], fragment=[], link=None, context=None, **variables):
        super(ShaderProgram, self).__init__(context=context)
        self._shaders = []
        self._variable_proxies = []

        self._delete_status = ProgramProxy(self._id, _gl.GL_DELETE_STATUS)
        self._link_status = ProgramProxy(self._id, _gl.GL_LINK_STATUS)
        self._validate_status = ProgramProxy(self._id, _gl.GL_VALIDATE_STATUS)
        self._info_log_length = ProgramProxy(self._id, _gl.GL_INFO_LOG_LENGTH)
        self._attached_shaders = ProgramProxy(self._id, _gl.GL_ATTACHED_SHADERS)
        self._active_attributes = ProgramProxy(self._id, _gl.GL_ACTIVE_ATTRIBUTES)
        self._active_attribute_max_length = ProgramProxy(self._id, _gl.GL_ACTIVE_ATTRIBUTE_MAX_LENGTH)
        self._active_uniforms = ProgramProxy(self._id, _gl.GL_ACTIVE_UNIFORMS)
        self._active_uniform_max_length = ProgramProxy(self._id, _gl.GL_ACTIVE_UNIFORM_MAX_LENGTH)
        self._program_binary_length = ProgramProxy(self._id, _gl.GL_PROGRAM_BINARY_LENGTH)
        self._transform_feedback_buffer_mode = ProgramProxy(self._id, _gl.GL_TRANSFORM_FEEDBACK_BUFFER_MODE, self.transform_feedback_buffer_modes)
        self._transform_feedback_varyings = ProgramProxy(self._id, _gl.GL_TRANSFORM_FEEDBACK_VARYINGS)
        self._transform_feedback_varying_max_length = ProgramProxy(self._id, _gl.GL_TRANSFORM_FEEDBACK_VARYING_MAX_LENGTH)

        shaders = list(shaders) if not isinstance(shaders, str) else [shaders]
        if not all(isinstance(x, Shader) for x in shaders):
            raise TypeError("expected Shader instance")
        shaders += [x if isinstance(x, VertexShader) else VertexShader(x, context=context)
                for x in (vertex if not isinstance(vertex, str) else [vertex])]
        shaders += [x if isinstance(x, TesselationControlShader) else TesselationControlShader(x, context=context)
                for x in (tess_control if not isinstance(tess_control, str) else [tess_control])]
        shaders += [x if isinstance(x, TesselationEvaluationShader) else TesselationEvaluationShader(x, context=context)
                for x in (tess_evaluation if not isinstance(tess_evaluation, str) else [tess_evaluation])]
        shaders += [x if isinstance(x, GeometryShader) else GeometryShader(x, context=context)
                for x in (geometry if not isinstance(geometry, str) else [geometry])]
        shaders += [x if isinstance(x, FragmentShader) else FragmentShader(x, context=context)
                for x in (fragment if not isinstance(fragment, str) else [fragment])]
        self.shaders.extend(shaders)

        if link is None:
            link = bool(shaders)
        if link:
            self.link()

        for name, proxy in list(self._get_active_attributes().items()):
            setattr(self, name, proxy)
            self._variable_proxies.append(proxy)
        for name, proxy in list(self._get_active_uniforms().items()):
            setattr(self, name, proxy)
            self._variable_proxies.append(proxy)

        self._frozen = True

        for key, value in list(variables.items()):
            setattr(self, key, value)

    def __setattr__(self, name, value):
        if self._frozen and not name.startswith("_") and not hasattr(self, name):
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))
        return super(ShaderProgram, self).__setattr__(name, value)

    def _on_bind(self):
        for proxy in self._variable_proxies:
            proxy._on_bind()

    def _on_release(self):
        for proxy in self._variable_proxies:
            proxy._on_release()

    def _attach(self, shader):
        with self._context:
            _gl.glAttachShader(self._id, shader._id)

    def _detach(self, shader):
        with self._context:
            _gl.glDetachShader(self._id, shader._id)

    def _get_texture_unit(self, name):
        candidates = [x for x in self._variable_proxies if x.name == name]
        if candidates:
            return candidates[0].__get__(self, get_texture_unit_instead_of_object=True)
        else:
            return None

    def _get_active_X(self, getter, location_getter, max_length, index):
        _size = _gl.GLint()
        _type = _gl.GLenum()
        _name = _gl.create_string_buffer(max_length)
        with self._context:
            getter(self._id, index, max_length, None, _gl.pointer(_size), _gl.pointer(_type), _name)
            location = location_getter(self._id, _name)
        return location, _size.value, ShaderDatatype._from_gl(_type.value), _name.value.decode("ascii")

    def _get_active_attribute(self, index):
        with self._context:
            return self._get_active_X(_gl.glGetActiveAttrib, _gl.glGetAttribLocation, self._active_attribute_max_length, index)

    def _get_active_attributes(self):
        return self._group_structs((self._get_active_attribute(i) for i in range(self._active_attributes)),
                Attribute, AttributeStruct, AttributeStructArray)

    def _get_active_uniform(self, index):
        with self._context:
            return self._get_active_X(_gl.glGetActiveUniform, _gl.glGetUniformLocation, self._active_uniform_max_length, index)

    def _get_active_uniforms(self):
        return self._group_structs((self._get_active_uniform(i) for i in range(self._active_uniforms)),
                Uniform, UniformStruct, UniformStructArray)

    def _group_structs(self, lst, make_variable, make_struct, coerce_array):
        array_of_structs_re = _re.compile("^([a-zA-Z_][a-zA-Z_0-9]*)\[([0-9]+)\]\.([a-zA-Z_][a-zA-Z_0-9]*)$")
        struct_re = _re.compile("^([a-zA-Z_][a-zA-Z_0-9]*)\.([a-zA-Z_][a-zA-Z_0-9]*)$")
        array_re = _re.compile("^([a-zA-Z_][a-zA-Z_0-9]*)\[([0-9]+)\]$")
        basic_re = _re.compile("^([a-zA-Z_][a-zA-Z_0-9]*)$")
        names = _odict()
        for location, size, dtype, name in lst:
            m = array_of_structs_re.match(name)
            if m is not None:
                name, index, field = m.groups()
                index = int(index)
                array = names.setdefault(name, coerce_array(name, parent=self))
                array[index, field] = make_variable(field, location, dtype, size, parent=self)
                continue
            m = struct_re.match(name)
            if m is not None:
                name, field = m.groups()
                struct = names.setdefault(name, make_struct(name, parent=self))
                struct[field] = make_variable(field, location, dtype, size, parent=self)
                continue
            m = array_re.match(name)
            if m is not None:
                name, index = m.groups()
                index = int(index)
                names[name] = make_variable(name, location, dtype, size, parent=self) # TODO how should this depend on index?
                continue
            m = basic_re.match(name)
            if m is not None:
                name, = m.groups()
                names[name] = make_variable(name, location, dtype, size, parent=self)
                continue
            raise NameError("shader variable '%s' could not be parsed" % name)
        return names

    @property
    def shaders(self):
        """The list of currently attached shaders.

        @rtype: L{ListProxy}
        """

        return ListProxy(self._shaders, self._attach, self._detach)

    def link(self):
        with self._context:
            old_program = self._context.current_program
            _gl.glLinkProgram(self._id) # linking binds the program without the context's knowledge

            # rebind the previous program circumventing any caching performed by the context
            if old_program is None:
                _gl.glUseProgram(0)
            elif old_program._id != self._id:
                _gl.glUseProgram(old_program._id)

            if self._link_status != _gl.GL_TRUE:
                raise ShaderLinkError(self._log)
            return self._log or None

    def validate(self):
        with self._context:
            self._on_bind()
            _gl.glValidateProgram(self._id)
            self._on_release()
            if self._validate_status != _gl.GL_TRUE:
                raise ShaderValidateError(self._log)
            return self._log or None

    def has_attribute_location(self, name):
        with self._context:
            return _gl.glGetAttribLocation(self._id, name.encode("ascii")) >= 0

    def get_attribute_location(self, name):
        with self._context:
            loc = _gl.glGetAttribLocation(self._id, name.encode("ascii"))
        if loc == -1:
            raise NameError("shader has no attribute '%s'" % name)
        return loc

    def has_frag_data_location(self, name):
        with self._context:
            return _gl.glGetFragDataLocation(self._id, name.encode("ascii")) >= 0

    def get_frag_data_location(self, name):
        with self._context:
            loc = _gl.glGetFragDataLocation(self._id, name.encode("ascii"))
        if loc == -1:
            raise NameError("shader has no frag data '%s'" % name)
        return loc

    @property
    def _log(self):
        """The current shader info log.

        @rtype: C{string}
        """

        _info_log = _gl.create_string_buffer(self._info_log_length)
        with self._context:
            _gl.glGetProgramInfoLog(self._id, self._info_log_length, _gl.POINTER(_gl.GLint)(), _info_log)
        return _info_log.value

__all__ = ["ShaderProgram"]




"""Descriptors for L{ShaderProgram} attributes.

End users should typically not need to use this module directly.

@bug: Attributes are currently unimplemented.

@author: Stephan Wenger
@date: 2012-02-29
"""

from collections import OrderedDict as _odict

class BaseAttribute(object):
    pass

class Attribute(BaseAttribute):
    def __init__(self, name, location, dtype, size, parent):
        self.name = name
        self.location = location
        self.dtype = dtype
        self.size = size
        self.parent = parent

    def __repr__(self):
        if self.size == 1:
            return "%s %s;" % (self.dtype, self.name)
        else:
            return "%s %s[%d];" % (self.dtype, self.name, self.size)

    def __str__(self):
        return "in %r" % self

    def __get__(self, obj, cls=None):
        """@todo: Implement this."""
        raise NotImplementedError

    def __set__(self, obj, value):
        """@todo: Implement this: if the shader is currently bound, set the attribute at once, else store only."""
        raise NotImplementedError

    def _on_bind(self):
        """@todo: Implement this: call C{glVertexAttrib} with C{index=self._location}."""
        pass

    def _on_release(self):
        """@todo: Implement this: restore old vertex attrib values if possible."""
        pass

class AttributeStruct(_odict, BaseAttribute):
    def __init__(self, name, parent):
        super(AttributeStruct, self).__init__()
        self.name = name
        self.parent = parent

    def __repr__(self):
        return "struct { %s } %s;" % (" ".join(str(value) for value in list(self.values())), self.name)

    def __str__(self):
        return "in %r" % self

    def __get__(self, obj, cls=None):
        """@todo: Implement this."""
        raise NotImplementedError

    def __set__(self, obj, value):
        """@todo: Implement this."""
        raise NotImplementedError

    def _on_bind(self):
        """@todo: Implement this: call C{glVertexAttrib} with C{index=self._location}."""
        pass

    def _on_release(self):
        """@todo: Implement this: restore old vertex attrib values if possible."""
        pass

class AttributeStructArray(_odict, BaseAttribute):
    def __init__(self, name, parent):
        super(AttributeStructArray, self).__init__()
        self.name = name
        self.parent = parent

    @property
    def size(self):
        return max(index for index, field in list(self.keys())) + 1

    def __repr__(self):
        unique_values = _odict((field, value) for ((index, field), value) in list(self.items()))
        return "struct { %s } %s[%d];" % (" ".join(repr(value) for value in list(unique_values.values())), self.name, self.size)

    def __str__(self):
        return "in %r" % self

    def __get__(self, obj, cls=None):
        """@todo: Implement this."""
        raise NotImplementedError

    def __set__(self, obj, value):
        """@todo: Implement this."""
        raise NotImplementedError

    def _on_bind(self):
        """@todo: Implement this: call C{glVertexAttrib} with C{index=self._location}."""
        pass

    def _on_release(self):
        """@todo: Implement this: restore old vertex attrib values if possible."""
        pass




#!/usr/bin/env python

"""Mandelbrot set renderer.

@author: Stephan Wenger
@date: 2012-03-23
"""

from numpy import linspace, array, minimum, maximum, cos, pi
from matplotlib import cm

from glitter import ShaderProgram, get_fullscreen_quad, Texture1D
from glitter.contexts.glut import GlutWindow, main_loop, get_elapsed_time
from glitter.raw import glut

vertex_shader = """
#version 400 core

layout(location=0) in vec4 in_position;
uniform vec2 minval, maxval;
out vec2 ex_texcoord;

void main() {
    gl_Position = in_position;
    ex_texcoord = (in_position.xy * 0.5 + 0.5) * (maxval - minval) + minval;
}
"""

fragment_shader = """
#version 400 core
#extension GL_ARB_texture_rectangle : enable

in vec2 ex_texcoord;
uniform sampler1D colormap;
layout(location=0) out vec4 out_color;

const int max_iteration = 1000;

void main() {
    vec2 xy = vec2(0.0, 0.0);
    int iteration = 0;
    while (xy.x * xy.x + xy.y * xy.y < 4 && iteration < max_iteration) {
        xy = vec2(xy.x * xy.x - xy.y * xy.y + ex_texcoord.x, 2 * xy.x * xy.y + ex_texcoord.y);
        iteration++;
    }
    float c = iteration - log(log(length(xy)));
    out_color = texture(colormap, 0.1 * c);
}
"""

class MandelbrotRenderer(object):
    transition_time = 0.3 # seconds
    update_step = 10 # milliseconds

    def __init__(self):
        self.window = GlutWindow(double=True, multisample=True)
        self.window.display_callback = self.display
        self.window.mouse_callback = self.mouse
        self.shader = ShaderProgram(vertex=vertex_shader, fragment=fragment_shader)
        self.shader.colormap = Texture1D(cm.spectral(linspace(0, 1, 256)), wrap_s="MIRRORED_REPEAT")
        self.shader.minval = (-2.5, -1.75)
        self.shader.maxval = (1.0, 1.75)
        self.vao = get_fullscreen_quad()
        self.history = []

    def display(self):
        self.window.clear()
        self.vao.draw()
        self.window.swap_buffers()

    def timer(self):
        t = min(1.0, (get_elapsed_time() - self.transition_start) / self.transition_time)
        x = 0.5 - 0.5 * cos(pi * t)
        self.shader.minval = self.minstart * (1 - x) + self.minend * x
        self.shader.maxval = self.maxstart * (1 - x) + self.maxend * x
        if t < 1:
            self.window.add_timer(self.update_step, self.timer)
        self.window.post_redisplay()

    def start_transition(self, minval, maxval):
        self.minstart, self.maxstart = self.shader.minval, self.shader.maxval
        self.minend, self.maxend = minval, maxval
        self.transition_start = get_elapsed_time()
        self.timer()

    def mouse(self, button, state, x, y):
        if button == glut.GLUT_LEFT_BUTTON:
            pos = array((x, self.window.shape[0] - y)) / array(self.window.shape[::-1], dtype=float)
            pos = pos * (self.shader.maxval - self.shader.minval) + self.shader.minval
            if state == glut.GLUT_DOWN:
                self.last_pos = pos
            elif state == glut.GLUT_UP and all(pos != self.last_pos):
                self.history.append((self.shader.minval, self.shader.maxval))
                self.start_transition(minimum(pos, self.last_pos), maximum(pos, self.last_pos))
        elif state == glut.GLUT_DOWN and button == glut.GLUT_RIGHT_BUTTON:
            if self.history:
                self.start_transition(*self.history.pop())

    def run(self):
        with self.shader:
            main_loop()

if __name__ == "__main__":
    MandelbrotRenderer().run()




"""Descriptors for per-drawbuffer state.

@author: Stephan Wenger
@date: 2012-02-29
"""

import itertools as _itertools
import numpy as _np

import glitter.raw as _gl
from glitter.utils import draw_buffers, bool8, GlitterError

class DrawBufferList(object):
    def __init__(self, _context):
        self._context = _context
        self._num_buffers = _context.max_draw_buffers
        if not 0 < self._num_buffers < 512: # sanity check
            raise GlitterError("implausible number of draw buffers detected; are you sure there is a current OpenGL context?")

    def __set__(self, obj, value):
        """Set all draw buffers.

        @param obj: Ignored.
        @type obj: any type
        @param value: The enum of the draw buffer to bind, a number if it is a color attachment.
        @type value: L{draw_buffers} or C{int}
        """

        _buffers = (_gl.GLenum * self._num_buffers)()
        for i, o in _itertools.islice(_itertools.izip_longest(list(range(self._num_buffers)), value, fillvalue=None), self._num_buffers):
            _buffers[i] = _gl.GL_NONE if o is None else _gl.GL_COLOR_ATTACHMENT0 + o if isinstance(o, int) else draw_buffers(o)._value
        with self._context:
            _gl.glDrawBuffers(self._num_buffers, _buffers)

    def __getitem__(self, index):
        """Get the draw buffer attached to this target.
        
        @param index: Index of the draw buffer to query.
        @type index: C{int}
        @return: The enum of the currently bound draw buffer enum, a number if it is a color attachment.
        @rtype: L{draw_buffers} or C{int}
        """

        if not 0 <= index < self._num_buffers:
            raise IndexError
        _buffer = _gl.GLint()
        with self._context:
            _gl.glGetIntegerv(_gl.GL_DRAW_BUFFER0 + index, _buffer)
        attachment = draw_buffers[_buffer.value]
        if attachment.name.startswith("COLOR_ATTACHMENT"):
            return int(attachment.lstrip("COLOR_ATTACHMENT", 1))
        else:
            return attachment

    def __setitem__(self, index, value):
        """Set a draw buffer.

        @param index: Index of the draw buffer to set.
        @type index: C{int}
        @param value: The enum of the draw buffer to bind, a number if it is a color attachment.
        @type value: L{draw_buffers} or C{int}
        """

        self.__set__(None, [value if i == index else self[i] for i in range(self._num_buffers)])

    def __delitem__(self, index):
        """Unbind a draw buffer.

        @param index: Index of the draw buffer to unbind.
        @type index: C{int}
        """

        self[index] = None

    def __len__(self):
        return self._num_buffers

    def __str__(self):
        return str(list(self))

    def __repr__(self):
        return str(self)

class ColorWritemaskList(object):
    def __init__(self, _context):
        self._context = _context
        self._num_buffers = _context.max_draw_buffers
        if not 0 < self._num_buffers < 512: # sanity check
            raise GlitterError("implausible number of draw buffers detected; are you sure there is a current OpenGL context?")

    def __set__(self, obj, value):
        if len(value) != self._num_buffers:
            raise ValueError("wrong size")
        for i, v in enumerate(value):
            self[i] = v

    def __getitem__(self, index):
        if not 0 <= index < self._num_buffers:
            raise IndexError
        mask = _np.empty(4, bool8.as_numpy())
        with self._context:
            _gl.glGetBooleani_v(_gl.GL_COLOR_WRITEMASK, index, _gl.cast(mask.ctypes, _gl.glGetBooleani_v.argtypes[-1]))
        return mask

    def __setitem__(self, index, value):
        with self._context:
            _gl.glColorMaski(index, *value)

    def __delitem__(self, index):
        self[index] = (True, True, True, True)

    def __len__(self):
        return self._num_buffers

    def __str__(self):
        return str(_np.array(self))

    def __repr__(self):
        return str(self)




"""Intuitive OpenGL wrappers.

Hacking glitter:
  - Make sure your class uses methods and properties of its L{Context} where
    possible instead of raw OpenGL calls; the context may cache some values or
    trigger callbacks, and calling OpenGL directly may break functionality in
    weird and unexpected ways. If you need to call raw GL functions, do so
    through the {glitter.raw} wrapper. If your raw GL invocations are generic
    enough, think about extending the L{Context} instead. Make sure to derive
    from L{GLObject} and use C{with self._context:} where necessary if your
    class does in any way interact directly with OpenGL. If your object is tied
    to a single context, derive from L{GLObject}. The constructor should take
    an optional C{context=None} parameter and pass it to L{GLObject.__init__}.
    Also be careful to create any additional OpenGL objects within the same
    context, preferably by passing the C{context} parameter to their
    constructors as well.
  - If your classes can be bound and unbound or change global state in some
    other way, they should act as context managers. Base classes are defined in
    the L{objects} module to help you doing it right.
  - Provide a convenient interface wherever possible. This includes casting
    values into the right types (e.g. to an appropriate array type using
    L{coerce_array}), failing gracefully with descriptive error messages, and
    adhering to conventions set by similar classes (e.g. returning C{self} from
    C{__enter__} methods, providing a C{__call__} method to set attributes by
    inheriting from L{StateMixin}).
  - If your class uses enums from the L{constants} module in a public
    interface, point a class variable of the same name to the enum so users do
    not need to import any constants manually. Functions that accept enums
    should cast input values into the appropriate constants using
    C{my_enum(value)} first to ensure type safety and to enable using enum
    constant names instead of L{EnumConstant}s.
  - Use absolute module names when importing other glitter modules. When
    importing from the same subpackage, use fully qualified names for the
    module (e.g. C{from glitter.utils.objects import GLObject}); otherwise,
    rely on the public interface of the subpackage defined in its
    C{__init__.py} and import from the subpackage directly (e.g. C{from
    glitter.utils import GLObject}). Do not import from the global C{glitter}
    module directly as visibility of subpackages may change.
  - Define the public interface of your module in C{__all__} and, if your
    module is meant for external use, C{import *} from your module in the
    C{__init__.py} of its parent. However, if your module has external
    dependencies other than numpy, do not C{import *} but have the user import
    your submodule manually.
  - Write docstrings for epydoc (use C{@todo}, C{@note}, C{@attention},
    C{@bug}, C{@warning} as appropriate).
  - Write tests for nosetests (in the C{tests} directory).
  - Write examples (in the C{examples} directory).

@todo: Use C{with self._context} for all GL calls, pass context when creating sub-objects.
@todo: Accept C{**kwargs} in C{__init__} for all classes with meaningful properties.
@todo: Use caching to reduce the number of bind and unbind operations caused by C{with} statements.
@todo: Implement NeHe tutorials as literate example programs, add an example for PyCUDA interoperability.
@todo: Write documentation and tests, expecially for using multiple objects at the same time (e.g., L{Texture}s, L{Context}s).
@todo: Provide example datasets for the examples (volume and mesh), link to source and to index for examples.
@todo: Provide a CUDA interoperability example.

@author: Stephan Wenger
@date: 2012-02-29
"""

from glitter.arrays import *
from glitter.contexts import *
from glitter.convenience import *
from glitter.framebuffers import *
from glitter.misc import *
from glitter.shaders import *
from glitter.textures import *
from glitter.utils import *

from glitter.raw import set_error_check, add_logger, GLError, LoggingWrapper, NamedConstant




"""Constants and enums.

@todo: Add stencil components to L{texture_formats}.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils.dtypes import uint8, uint16, uint32, int8, int16, int32, float32
from glitter.utils.enum import Enum

texture_formats = [
        ((float32, 1), _gl.GL_DEPTH_COMPONENT32F, (_gl.GL_FLOAT, _gl.GL_DEPTH_COMPONENT)),
        ((int16, 1), _gl.GL_DEPTH_COMPONENT16, (_gl.GL_SHORT, _gl.GL_DEPTH_COMPONENT)),
        ((int32, 1), _gl.GL_DEPTH_COMPONENT24, (_gl.GL_INT, _gl.GL_DEPTH_COMPONENT)),
        ((int32, 1), _gl.GL_DEPTH_COMPONENT32, (_gl.GL_INT, _gl.GL_DEPTH_COMPONENT)),

        ((uint8,   1), _gl.GL_R8UI,     (_gl.GL_UNSIGNED_BYTE,  _gl.GL_RED_INTEGER )),
        ((int8,    1), _gl.GL_R8I,      (_gl.GL_BYTE,           _gl.GL_RED_INTEGER )),
        ((uint16,  1), _gl.GL_R16UI,    (_gl.GL_UNSIGNED_SHORT, _gl.GL_RED_INTEGER )),
        ((int16,   1), _gl.GL_R16I,     (_gl.GL_SHORT,          _gl.GL_RED_INTEGER )),
        ((uint32,  1), _gl.GL_R32UI,    (_gl.GL_UNSIGNED_INT,   _gl.GL_RED_INTEGER )),
        ((int32,   1), _gl.GL_R32I,     (_gl.GL_INT,            _gl.GL_RED_INTEGER )),
        ((float32, 1), _gl.GL_R32F,     (_gl.GL_FLOAT,          _gl.GL_RED         )),
        ((uint8,   2), _gl.GL_RG8UI,    (_gl.GL_UNSIGNED_BYTE,  _gl.GL_RG_INTEGER  )),
        ((int8,    2), _gl.GL_RG8I,     (_gl.GL_BYTE,           _gl.GL_RG_INTEGER  )),
        ((uint16,  2), _gl.GL_RG16UI,   (_gl.GL_UNSIGNED_SHORT, _gl.GL_RG_INTEGER  )),
        ((int16,   2), _gl.GL_RG16I,    (_gl.GL_SHORT,          _gl.GL_RG_INTEGER  )),
        ((uint32,  2), _gl.GL_RG32UI,   (_gl.GL_UNSIGNED_INT,   _gl.GL_RG_INTEGER  )),
        ((int32,   2), _gl.GL_RG32I,    (_gl.GL_INT,            _gl.GL_RG_INTEGER  )),
        ((float32, 2), _gl.GL_RG32F,    (_gl.GL_FLOAT,          _gl.GL_RG          )),
        ((uint8,   3), _gl.GL_RGB8UI,   (_gl.GL_UNSIGNED_BYTE,  _gl.GL_RGB_INTEGER )),
        ((int8,    3), _gl.GL_RGB8I,    (_gl.GL_BYTE,           _gl.GL_RGB_INTEGER )),
        ((uint16,  3), _gl.GL_RGB16UI,  (_gl.GL_UNSIGNED_SHORT, _gl.GL_RGB_INTEGER )),
        ((int16,   3), _gl.GL_RGB16I,   (_gl.GL_SHORT,          _gl.GL_RGB_INTEGER )),
        ((uint32,  3), _gl.GL_RGB32UI,  (_gl.GL_UNSIGNED_INT,   _gl.GL_RGB_INTEGER )),
        ((int32,   3), _gl.GL_RGB32I,   (_gl.GL_INT,            _gl.GL_RGB_INTEGER )),
        ((float32, 3), _gl.GL_RGB32F,   (_gl.GL_FLOAT,          _gl.GL_RGB         )),
        ((uint8,   4), _gl.GL_RGBA8UI,  (_gl.GL_UNSIGNED_BYTE,  _gl.GL_RGBA_INTEGER)),
        ((int8,    4), _gl.GL_RGBA8I,   (_gl.GL_BYTE,           _gl.GL_RGBA_INTEGER)),
        ((uint16,  4), _gl.GL_RGBA16UI, (_gl.GL_UNSIGNED_SHORT, _gl.GL_RGBA_INTEGER)),
        ((int16,   4), _gl.GL_RGBA16I,  (_gl.GL_SHORT,          _gl.GL_RGBA_INTEGER)),
        ((uint32,  4), _gl.GL_RGBA32UI, (_gl.GL_UNSIGNED_INT,   _gl.GL_RGBA_INTEGER)),
        ((int32,   4), _gl.GL_RGBA32I,  (_gl.GL_INT,            _gl.GL_RGBA_INTEGER)),
        ((float32, 4), _gl.GL_RGBA32F,  (_gl.GL_FLOAT,          _gl.GL_RGBA        )),
        ]
"""Mapping between C{numpy} types and OpenGL types.

First column: C{numpy} datatype, number of color channels
Second column: OpenGL internal format
Third column: OpenGL type, OpenGL format
"""

dtype_to_gl_iformat =   dict((x[0],    x[1]   ) for x in texture_formats)
gl_iformat_to_dtype =   dict((x[1],    x[0]   ) for x in texture_formats)
dtype_to_gl_format =    dict((x[0],    x[2][1]) for x in texture_formats)
gl_format_to_dtype =    dict((x[2][1], x[0]   ) for x in texture_formats)
gl_iformat_to_gl_type = dict((x[1],    x[2][0]) for x in texture_formats)

texture_compare_funcs = Enum(
        LEQUAL=_gl.GL_LEQUAL,
        GEQUAL=_gl.GL_GEQUAL,
        LESS=_gl.GL_LESS,
        GREATER=_gl.GL_GREATER,
        EQUAL=_gl.GL_EQUAL,
        NOTEQUAL=_gl.GL_NOTEQUAL,
        ALWAYS=_gl.GL_ALWAYS,
        NEVER=_gl.GL_NEVER,
        )

texture_compare_modes = Enum(
        COMPARE_REF_TO_TEXTURE=_gl.GL_COMPARE_REF_TO_TEXTURE,
        NONE=_gl.GL_NONE,
        )

texture_min_filters = Enum(
        NEAREST=_gl.GL_NEAREST,
        LINEAR=_gl.GL_LINEAR,
        NEAREST_MIPMAP_NEAREST=_gl.GL_NEAREST_MIPMAP_NEAREST,
        LINEAR_MIPMAP_NEAREST=_gl.GL_LINEAR_MIPMAP_NEAREST,
        NEAREST_MIPMAP_LINEAR=_gl.GL_NEAREST_MIPMAP_LINEAR,
        LINEAR_MIPMAP_LINEAR=_gl.GL_LINEAR_MIPMAP_LINEAR,
        )

texture_mag_filters = Enum(
        NEAREST=_gl.GL_NEAREST,
        LINEAR=_gl.GL_LINEAR,
        )

texture_swizzles = Enum(
        RED=_gl.GL_RED,
        GREEN=_gl.GL_GREEN,
        BLUE=_gl.GL_BLUE,
        ALPHA=_gl.GL_ALPHA,
        ZERO=_gl.GL_ZERO,
        ONE=_gl.GL_ONE,
        )

texture_wrapmodes = Enum(
        CLAMP_TO_EDGE=_gl.GL_CLAMP_TO_EDGE,
        CLAMP_TO_BORDER=_gl.GL_CLAMP_TO_BORDER,
        MIRRORED_REPEAT=_gl.GL_MIRRORED_REPEAT,
        REPEAT=_gl.GL_REPEAT,
        )

primitive_types = Enum(
        POINTS=_gl.GL_POINTS,
        LINE_STRIP=_gl.GL_LINE_STRIP,
        LINE_LOOP=_gl.GL_LINE_LOOP,
        LINES=_gl.GL_LINES,
        LINE_STRIP_ADJACENCY=_gl.GL_LINE_STRIP_ADJACENCY,
        LINES_ADJACENCY=_gl.GL_LINES_ADJACENCY,
        TRIANGLE_STRIP=_gl.GL_TRIANGLE_STRIP,
        TRIANGLE_FAN=_gl.GL_TRIANGLE_FAN,
        TRIANGLES=_gl.GL_TRIANGLES,
        TRIANGLE_STRIP_ADJACENCY=_gl.GL_TRIANGLE_STRIP_ADJACENCY,
        TRIANGLES_ADJACENCY=_gl.GL_TRIANGLES_ADJACENCY,
        PATCHES=_gl.GL_PATCHES,
        )

buffer_usages = Enum(
        STREAM_DRAW=_gl.GL_STREAM_DRAW,
        STREAM_READ=_gl.GL_STREAM_READ,
        STREAM_COPY=_gl.GL_STREAM_COPY,
        STATIC_DRAW=_gl.GL_STATIC_DRAW,
        STATIC_READ=_gl.GL_STATIC_READ,
        STATIC_COPY=_gl.GL_STATIC_COPY,
        DYNAMIC_DRAW=_gl.GL_DYNAMIC_DRAW,
        DYNAMIC_READ=_gl.GL_DYNAMIC_READ,
        DYNAMIC_COPY=_gl.GL_DYNAMIC_COPY,
        )

buffer_dimensions_to_primitive = {1: primitive_types.POINTS, 2: primitive_types.LINES, 3: primitive_types.TRIANGLES}
primitive_to_buffer_dimensions = {primitive_types.POINTS: 1, primitive_types.LINES: 2, primitive_types.TRIANGLES: 3}

blend_functions = Enum(
        ZERO=_gl.GL_ZERO,
        ONE=_gl.GL_ONE,
        SRC_COLOR=_gl.GL_SRC_COLOR,
        ONE_MINUS_SRC_COLOR=_gl.GL_ONE_MINUS_SRC_COLOR,
        DST_COLOR=_gl.GL_DST_COLOR,
        ONE_MINUS_DST_COLOR=_gl.GL_ONE_MINUS_DST_COLOR,
        SRC_ALPHA=_gl.GL_SRC_ALPHA,
        ONE_MINUS_SRC_ALPHA=_gl.GL_ONE_MINUS_SRC_ALPHA,
        DST_ALPHA=_gl.GL_DST_ALPHA,
        ONE_MINUS_DST_ALPHA=_gl.GL_ONE_MINUS_DST_ALPHA,
        CONSTANT_COLOR=_gl.GL_CONSTANT_COLOR,
        ONE_MINUS_CONSTANT_COLOR=_gl.GL_ONE_MINUS_CONSTANT_COLOR,
        CONSTANT_ALPHA=_gl.GL_CONSTANT_ALPHA,
        ONE_MINUS_CONSTANT_ALPHA=_gl.GL_ONE_MINUS_CONSTANT_ALPHA,
        SRC_ALPHA_SATURATE=_gl.GL_SRC_ALPHA_SATURATE,
)

blend_equations = Enum(
        ADD=_gl.GL_FUNC_ADD,
        SUBTRACT=_gl.GL_FUNC_SUBTRACT,
        REVERSE_SUBTRACT=_gl.GL_FUNC_REVERSE_SUBTRACT,
        MIN=_gl.GL_MIN,
        MAX=_gl.GL_MAX,
)

depth_functions = Enum(
        NEVER=_gl.GL_NEVER,
        LESS=_gl.GL_LESS,
        EQUAL=_gl.GL_EQUAL,
        LEQUAL=_gl.GL_LEQUAL,
        GREATER=_gl.GL_GREATER,
        NOTEQUAL=_gl.GL_NOTEQUAL,
        GEQUAL=_gl.GL_GEQUAL,
        ALWAYS=_gl.GL_ALWAYS,
)

draw_buffers = Enum(
        NONE=_gl.GL_NONE,
        FRONT_LEFT=_gl.GL_FRONT_LEFT,
        FRONT_RIGHT=_gl.GL_FRONT_RIGHT,
        BACK_LEFT=_gl.GL_BACK_LEFT,
        BACK_RIGHT=_gl.GL_BACK_RIGHT,
        FRONT=_gl.GL_FRONT,
        BACK=_gl.GL_BACK,
        LEFT=_gl.GL_LEFT,
        RIGHT=_gl.GL_RIGHT,
        FRONT_AND_BACK=_gl.GL_FRONT_AND_BACK,
)
for key, value in list(_gl.__dict__.items()):
    if key.startswith("GL_COLOR_ATTACHMENT"):
        draw_buffers._add(key[3:], value)

hints = Enum(
        FASTEST=_gl.GL_FASTEST,
        NICEST=_gl.GL_NICEST,
        DONT_CARE=_gl.GL_DONT_CARE,
)

provoking_vertices = Enum(
        PROVOKING=_gl.GL_PROVOKING_VERTEX,
        FIRST=_gl.GL_FIRST_VERTEX_CONVENTION,
        LAST=_gl.GL_LAST_VERTEX_CONVENTION,
        UNDEFINED=_gl.GL_UNDEFINED_VERTEX,
)

logic_op_modes = Enum(
        CLEAR=_gl.GL_CLEAR,
        SET=_gl.GL_SET,
        COPY=_gl.GL_COPY,
        COPY_INVERTED=_gl.GL_COPY_INVERTED,
        NOOP=_gl.GL_NOOP,
        INVERT=_gl.GL_INVERT,
        AND=_gl.GL_AND,
        NAND=_gl.GL_NAND,
        OR=_gl.GL_OR,
        NOR=_gl.GL_NOR,
        XOR=_gl.GL_XOR,
        EQUIV=_gl.GL_EQUIV,
        AND_REVERSE=_gl.GL_AND_REVERSE,
        AND_INVERTED=_gl.GL_AND_INVERTED,
        OR_REVERSE=_gl.GL_OR_REVERSE,
        OR_INVERTED=_gl.GL_OR_INVERTED,
)

provoke_modes = Enum(
        FIRST_VERTEX_CONVENTION=_gl.GL_FIRST_VERTEX_CONVENTION,
        LAST_VERTEX_CONVENTION=_gl.GL_LAST_VERTEX_CONVENTION,
)

color_read_formats = Enum(
        STENCIL_INDEX=_gl.GL_STENCIL_INDEX,
        DEPTH_COMPONENT=_gl.GL_DEPTH_COMPONENT,
        DEPTH_STENCIL=_gl.GL_DEPTH_STENCIL,
        RED=_gl.GL_RED,
        GREEN=_gl.GL_GREEN,
        BLUE=_gl.GL_BLUE,
        RGB=_gl.GL_RGB,
        BGR=_gl.GL_BGR,
        RGBA=_gl.GL_RGBA,
        BGRA=_gl.GL_BGRA,
)

color_read_types = Enum(
        UNSIGNED_BYTE=_gl.GL_UNSIGNED_BYTE,
        BYTE=_gl.GL_BYTE,
        UNSIGNED_SHORT=_gl.GL_UNSIGNED_SHORT,
        SHORT=_gl.GL_SHORT,
        UNSIGNED_INT=_gl.GL_UNSIGNED_INT,
        INT=_gl.GL_INT,
        HALF_FLOAT=_gl.GL_HALF_FLOAT,
        FLOAT=_gl.GL_FLOAT,
        UNSIGNED_BYTE_3_3_2=_gl.GL_UNSIGNED_BYTE_3_3_2,
        UNSIGNED_BYTE_2_3_3_REV=_gl.GL_UNSIGNED_BYTE_2_3_3_REV,
        UNSIGNED_SHORT_5_6_5=_gl.GL_UNSIGNED_SHORT_5_6_5,
        UNSIGNED_SHORT_5_6_5_REV=_gl.GL_UNSIGNED_SHORT_5_6_5_REV,
        UNSIGNED_SHORT_4_4_4_4=_gl.GL_UNSIGNED_SHORT_4_4_4_4,
        UNSIGNED_SHORT_4_4_4_4_REV=_gl.GL_UNSIGNED_SHORT_4_4_4_4_REV,
        UNSIGNED_SHORT_5_5_5_1=_gl.GL_UNSIGNED_SHORT_5_5_5_1,
        UNSIGNED_SHORT_1_5_5_5_REV=_gl.GL_UNSIGNED_SHORT_1_5_5_5_REV,
        UNSIGNED_INT_8_8_8_8=_gl.GL_UNSIGNED_INT_8_8_8_8,
        UNSIGNED_INT_8_8_8_8_REV=_gl.GL_UNSIGNED_INT_8_8_8_8_REV,
        UNSIGNED_INT_10_10_10_2=_gl.GL_UNSIGNED_INT_10_10_10_2,
        UNSIGNED_INT_2_10_10_10_REV=_gl.GL_UNSIGNED_INT_2_10_10_10_REV,
        UNSIGNED_INT_24_8=_gl.GL_UNSIGNED_INT_24_8,
        UNSIGNED_INT_10F_11F_11F_REV=_gl.GL_UNSIGNED_INT_10F_11F_11F_REV,
        UNSIGNED_INT_5_9_9_9_REV=_gl.GL_UNSIGNED_INT_5_9_9_9_REV,
        FLOAT_32_UNSIGNED_INT_24_8_REV=_gl.GL_FLOAT_32_UNSIGNED_INT_24_8_REV,
)

read_buffers = Enum(
        NONE=_gl.GL_NONE,
        FRONT_LEFT=_gl.GL_FRONT_LEFT,
        FRONT_RIGHT=_gl.GL_FRONT_RIGHT,
        BACK_LEFT=_gl.GL_BACK_LEFT,
        BACK_RIGHT=_gl.GL_BACK_RIGHT,
        FRONT=_gl.GL_FRONT,
        BACK=_gl.GL_BACK,
        LEFT=_gl.GL_LEFT,
        RIGHT=_gl.GL_RIGHT,
)
for key, value in list(_gl.__dict__.items()):
    if key.startswith("GL_COLOR_ATTACHMENT"):
        read_buffers._add(key[3:], value)

transform_feedback_buffer_modes = Enum(
        SEPARATE_ATTRIBS=_gl.GL_SEPARATE_ATTRIBS,
        INTERLEAVED_ATTRIBS=_gl.GL_INTERLEAVED_ATTRIBS,
)

client_wait_sync_returns = Enum(
        ALREADY_SIGNALED=_gl.GL_ALREADY_SIGNALED,
        TIMEOUT_EXPIRED=_gl.GL_TIMEOUT_EXPIRED,
        CONDITION_SATISFIED=_gl.GL_CONDITION_SATISFIED,
        WAIT_FAILED=_gl.GL_WAIT_FAILED,
)

framebuffer_status = Enum(
    COMPLETE=_gl.GL_FRAMEBUFFER_COMPLETE,
    UNDEFINED=_gl.GL_FRAMEBUFFER_UNDEFINED,
    UNSUPPORTED=_gl.GL_FRAMEBUFFER_UNSUPPORTED,
    INCOMPLETE_ATTACHMENT=_gl.GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT,
    INCOMPLETE_MISSING_ATTACHMENT=_gl.GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT,
    INCOMPLETE_DRAW_BUFFER=_gl.GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER,
    INCOMPLETE_READ_BUFFER=_gl.GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER,
    INCOMPLETE_MULTISAMPLE=_gl.GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE,
    INCOMPLETE_LAYER_TARGETS=_gl.GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS,
)

cull_face_modes = Enum(
    FRONT=_gl.GL_FRONT,
    BACK=_gl.GL_BACK,
    FRONT_AND_BACK=_gl.GL_FRONT_AND_BACK,
)

front_face_modes = Enum(
    CW=_gl.GL_CW,
    CCW=_gl.GL_CCW,
)

polygon_modes = Enum(
    POINT=_gl.GL_POINT,
    LINE=_gl.GL_LINE,
    FILL=_gl.GL_FILL,
)

__all__ = [
    "texture_formats",
    "dtype_to_gl_iformat",
    "gl_iformat_to_dtype",
    "dtype_to_gl_format",
    "gl_format_to_dtype",
    "gl_iformat_to_gl_type",
    "texture_compare_funcs",
    "texture_compare_modes",
    "texture_min_filters",
    "texture_mag_filters",
    "texture_swizzles",
    "texture_wrapmodes",
    "primitive_types",
    "buffer_usages",
    "buffer_dimensions_to_primitive",
    "primitive_to_buffer_dimensions",
    "blend_functions",
    "blend_equations",
    "depth_functions",
    "draw_buffers",
    "hints",
    "provoking_vertices",
    "logic_op_modes",
    "provoke_modes",
    "color_read_formats",
    "color_read_types",
    "read_buffers",
    "transform_feedback_buffer_modes",
    "client_wait_sync_returns",
    "framebuffer_status",
    "cull_face_modes",
    "front_face_modes",
    "polygon_modes",
]




"""Sampler class.

@todo: Implement getters and setters for C{Sampler.unit}, but take care not to
confuse any existing bindings on the stack. This problem is very similar to
rebinding L{buffers<BaseBuffer>} to different targets and should probably be
handled in a similar way.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils import texture_compare_funcs, texture_compare_modes, texture_min_filters, texture_mag_filters, texture_wrapmodes, ManagedObject, BindableObject

class Sampler(ManagedObject, BindableObject):
    _generate_id = _gl.glGenSamplers
    _delete_id = _gl.glDeleteSamplers
    _db = "samplers"
    _binding = "sampler_binding"

    compare_funcs = texture_compare_funcs
    compare_modes = texture_compare_modes
    min_filters = texture_min_filters
    mag_filters = texture_mag_filters
    wrapmodes = texture_wrapmodes

    def __init__(self, unit, context=None):
        super(Sampler, self).__init__(context=context)
        if not 0 <= unit < self._context.max_combined_texture_image_units:
            raise ValueError("invalid unit: %s" % unit)
        self._target = unit

    @property
    def compare_func(self):
        _compare_func = _gl.GLint()
        _gl.glGetSamplerParameteriv(self._id, _gl.GL_TEXTURE_COMPARE_FUNC, _gl.pointer(_compare_func))
        return self.compare_funcs[_compare_func.value]

    @compare_func.setter
    def compare_func(self, compare_func):
        _compare_func = _gl.GLint(self.compare_funcs(compare_func)._value)
        _gl.glSamplerParameteriv(self._id, _gl.GL_TEXTURE_COMPARE_FUNC, _gl.pointer(_compare_func))

    @property
    def compare_mode(self):
        _compare_mode = _gl.GLint()
        _gl.glGetSamplerParameteriv(self._id, _gl.GL_TEXTURE_COMPARE_MODE, _gl.pointer(_compare_mode))
        return self.compare_modes[_compare_mode.value]

    @compare_mode.setter
    def compare_mode(self, compare_mode):
        _compare_mode = _gl.GLint(self.compare_modes(compare_mode)._value)
        _gl.glSamplerParameteriv(self._id, _gl.GL_TEXTURE_COMPARE_MODE, _gl.pointer(_compare_mode))
    
    @property
    def lod_bias(self):
        _lod_bias = _gl.GLfloat()
        _gl.glGetSamplerParameterfv(self._id, _gl.GL_TEXTURE_LOD_BIAS, _gl.pointer(_lod_bias))
        return _lod_bias.value

    @lod_bias.setter
    def lod_bias(self, lod_bias):
        _gl.glSamplerParameterfv(self._id, _gl.GL_TEXTURE_LOD_BIAS, _gl.pointer(_gl.GLfloat(lod_bias)))

    @property
    def min_filter(self):
        _min_filter = _gl.GLint()
        _gl.glGetSamplerParameteriv(self._id, _gl.GL_TEXTURE_MIN_FILTER, _gl.pointer(_min_filter))
        return self.min_filters[_min_filter.value]

    @min_filter.setter
    def min_filter(self, min_filter):
        _min_filter = _gl.GLint(self.min_filters(min_filter)._value)
        _gl.glSamplerParameteriv(self._id, _gl.GL_TEXTURE_MIN_FILTER, _gl.pointer(_min_filter))

    @property
    def mag_filter(self):
        _mag_filter = _gl.GLint()
        _gl.glGetSamplerParameteriv(self._id, _gl.GL_TEXTURE_MAG_FILTER, _gl.pointer(_mag_filter))
        return self.mag_filters[_mag_filter.value]

    @mag_filter.setter
    def mag_filter(self, mag_filter):
        _mag_filter = _gl.GLint(self.mag_filters(mag_filter)._value)
        _gl.glSamplerParameteriv(self._id, _gl.GL_TEXTURE_MAG_FILTER, _gl.pointer(_mag_filter))

    @property
    def min_lod(self):
        _min_lod = _gl.GLint()
        _gl.glGetSamplerParameteriv(self._id, _gl.GL_TEXTURE_MIN_LOD, _gl.pointer(_min_lod))
        return _min_lod.value

    @min_lod.setter
    def min_lod(self, min_lod):
        _gl.glSamplerParameteriv(self._id, _gl.GL_TEXTURE_MIN_LOD, _gl.pointer(_gl.GLint(min_lod)))

    @property
    def max_lod(self):
        _max_lod = _gl.GLint()
        _gl.glGetSamplerParameteriv(self._id, _gl.GL_TEXTURE_MAX_LOD, _gl.pointer(_max_lod))
        return _max_lod.value

    @max_lod.setter
    def max_lod(self, max_lod):
        _gl.glSamplerParameteriv(self._id, _gl.GL_TEXTURE_MAX_LOD, _gl.pointer(_gl.GLint(max_lod)))

    @property
    def wrap_s(self):
        _wrap_s = _gl.GLint()
        _gl.glGetSamplerParameteriv(self._id, _gl.GL_TEXTURE_WRAP_S, _gl.pointer(_wrap_s))
        return self.wrapmodes[_wrap_s.value]

    @wrap_s.setter
    def wrap_s(self, wrap_s):
        _wrap_s = _gl.GLint(self.wrapmodes(wrap_s)._value)
        _gl.glSamplerParameteriv(self._id, _gl.GL_TEXTURE_WRAP_S, _gl.pointer(_wrap_s))

    @property
    def wrap_t(self):
        _wrap_t = _gl.GLint()
        _gl.glGetSamplerParameteriv(self._id, _gl.GL_TEXTURE_WRAP_T, _gl.pointer(_wrap_t))
        return self.wrapmodes[_wrap_t.value]

    @wrap_t.setter
    def wrap_t(self, wrap_t):
        _wrap_t = _gl.GLint(self.wrapmodes(wrap_t)._value)
        _gl.glSamplerParameteriv(self._id, _gl.GL_TEXTURE_WRAP_T, _gl.pointer(_wrap_t))

    @property
    def wrap_r(self):
        _wrap_r = _gl.GLint()
        _gl.glGetSamplerParameteriv(self._id, _gl.GL_TEXTURE_WRAP_R, _gl.pointer(_wrap_r))
        return self.wrapmodes[_wrap_r.value]

    @wrap_r.setter
    def wrap_r(self, wrap_r):
        _wrap_r = _gl.GLint(self.wrapmodes(wrap_r)._value)
        _gl.glSamplerParameteriv(self._id, _gl.GL_TEXTURE_WRAP_R, _gl.pointer(_wrap_r))

__all__ = ["Sampler"]




"""Vertex array class.

@todo: Rethink vertex array drawing: e.g., differing number of elements in vertex and color buffer, allow for use of C{glVertexAttribDivisor}.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils import ManagedObject, BindableObject
from glitter.arrays.arraybuffer import ArrayBuffer
from glitter.arrays.elementarray import ElementArrayBuffer

class VertexArray(ManagedObject, BindableObject):
    _generate_id = _gl.glGenVertexArrays
    _delete_id = _gl.glDeleteVertexArrays
    _db = "vertex_arrays"
    _binding = "vertex_array_binding"

    def __init__(self, *attributes, **kwargs):
        """Create a new vertex array.

        @param attributes: Buffers to bind to vertex attributes.
        @type attributes: C{list} of L{ArrayBuffer}s or C{numpy.ndarray}s
        @param kwargs: Named arguments.
        @type kwargs: C{dict}
        @keyword context: The context in which to create the vertex array.
        @type context: L{Context}
        @keyword elements: A buffer containing the element indices.
        @type elements: L{ElementArrayBuffer} or C{numpy.ndarray}
        """

        super(VertexArray, self).__init__(context=kwargs.pop("context", None))
        self._attributes = {}

        if isinstance(attributes, dict):
            attributes = dict(attributes)
        else:
            attributes = dict(enumerate(attributes))
        for i in range(self._context.max_vertex_attribs):
            self[i] = attributes.pop(i, None)
        if attributes:
            raise ValueError("vertex array has no attribute(s) %s" % ", ".join("'%s'" % x for x in list(attributes.keys())))
        
        self.elements = kwargs.pop("elements", None)

        if kwargs:
            raise TypeError("__init__() got an unexpected keyword argument '%s'" % list(kwargs.keys())[0])

    def __getitem__(self, index):
        return self._attributes[index]

    def __setitem__(self, index, value):
        if value is None:
            with self:
                _gl.glVertexAttribPointer(index, 4, _gl.GL_FLOAT, _gl.GL_FALSE, 0, _gl.POINTER(_gl.GLvoid)())
                _gl.glDisableVertexAttribArray(index)
        else:
            if not isinstance(value, ArrayBuffer):
                value = ArrayBuffer(value, context=self._context)
            with self:
                with value:
                    value._use(index)
                _gl.glEnableVertexAttribArray(index)
        self._attributes[index] = value

    def __delitem__(self, index):
        self[index] = None

    @property
    def elements(self):
        return self._elements

    @elements.setter
    def elements(self, elements):
        if elements is not None and not isinstance(elements, ElementArrayBuffer):
            elements = ElementArrayBuffer(elements, context=self._context)
        with self:
            self._context.element_array_buffer_binding = elements
        self._elements = elements

    @elements.deleter
    def elements(self):
        self.elements = None

    def draw(self, mode=None, count=None, first=0, instances=None, index=None):
        with self:
            if index is None:
                if self.elements is not None:
                    self.elements.draw(mode, count, first, instances)
                else:
                    min(x for x in list(self._attributes.items()) if x[1] is not None)[1].draw(mode, count, first, instances)
            else:
                self[index].draw(mode, count, first, instances)

__all__ = ["VertexArray"]




#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

#! This file is a literate Python program. You can compile the documentation
#! using mylit (http://pypi.python.org/pypi/mylit/).
## title = "glitter Example: Qt"
## stylesheet = "pygments_style.css"

# <h1><i>glitter</i> Example: Qt</h1>

# <h2>Summary</h2>

# This program will show a Qt application to display meshes.

# <img src="qt.png">

# <h2>Front matter</h2>

# <h3>Module docstring</h3>

# The module docstring is used as a description of this example in the
# generated documentation:
"""Example for Qt (PySide) interaction.

@author: Stephan Wenger
@date: 2012-03-16
"""

# <h3>Imports</h3>
from __future__ import unicode_literals

# Some math functions are required for mouse interaction:
from math import sqrt, exp

# We use <a href="http://www.pyside.org/">PySide</a> for <a
# href="http://qt.nokia.com/">Qt</a> interaction:
from PySide import QtCore, QtGui

# Note that we did not import the <code>QtOpenGL</code> package. Instead, we
# will use a Qt OpenGL widget that doubles as a <i>glitter</i>
# <code>Context</code>:
from glitter.contexts.qt import QtWidget
# If you'd rather use an existing Qt OpenGL context,
# <code>QtContextWrapper</code> provides an alternative solution.

# Since there are no more build-in matrices in the OpenGL core profile, we use
# <i>glitter</i>'s matrix constructors. <i>glitter</i> also provides a method
# for loading meshes from files that we will use:
from glitter import rotation_matrix, scale_matrix, identity_matrix, load_mesh

# <h2>Qt GUI</h2>

# <h3>Qt OpenGL Widget</h3>

# OpenGL rendering in Qt is through an OpenGL widget. Instead of subclassing
# <code>QGLWidget</code> directly, we inherit from <i>glitter</i>'s
# <code>QtWidget</code> which acts as a <i>glitter</i> <code>Context</code>.
# This not only provides us with a convenient interface for changing OpenGL
# state, it also ensures that the correct context is made active whenever
# OpenGL commands are issued.
class Canvas(QtWidget):
    # <h4>Initialization</h4>
    # The canvas will store a mesh to display, a modelview matrix that can be
    # changed by moving the mouse, and a helper variable for the mouse
    # interaction:
    def __init__(self, parent=None):
        super(Canvas, self).__init__(parent)
        self.mesh = None
        self.modelview_matrix = identity_matrix()
        self.lastPos = QtCore.QPoint()

    # The size of the canvas is specified by overriding <code>sizeHint()</code>:
    def sizeHint(self):
        return QtCore.QSize(512, 512)

    # When the user requests loading a mesh via the menu system, we create a
    # <code>Pipeline</code> containing the mesh vertices and an appropriate
    # shader, all with a single call to <code>load_mesh()</code> from the
    # <code>glitter.convenience</code> module. Also, we can pass additional
    # parameters to the <code>Pipeline</code> constructor. In this case, we
    # want depth testing to be enabled whenever the pipeline is executed, so we
    # pass <code>depth_text=True</code>:
    def loadMesh(self, filename):
        self.mesh = load_mesh(filename, context=self, depth_test=True)
        # Note that we pass the canvas as the <code>context</code> parameter to
        # make sure the pipeline is created in the correct context. This is
        # necessary because in a real-world application, we cannot be sure that
        # the canvas context is currently active.

        # When the mesh is loaded, we ask Qt to redraw the OpenGL screen:
        self.updateGL()

    # <h4>Mouse interaction</h4>
    # The viewpoint can be changed by moving the mouse with a button pressed:
    # left button rotates, right button zooms.

    # When a mouse button is pressed, we store the position where it was
    # pressed:
    def mousePressEvent(self, event):
        self.lastPos = QtCore.QPoint(event.pos())

    # When the mouse is moved, we take action according to the type of the
    # pressed button:
    def mouseMoveEvent(self, event):
        # First, we compute the mouse movement since the last time we processed
        # a mouse event. If there was no movement (which does happen from time
        # to time), we exit early to avoid division by zero later on:
        dx, dy = event.x() - self.lastPos.x(), event.y() - self.lastPos.y()
        if dx == dy == 0:
            return

        # If the left mouse button is down, we rotate the modelview matrix
        # about an axis within the image plane that is perpendicular to the
        # direction of mouse movement. The amount of rotation is proportional
        # to the distance travelled. Then we cause a screen redraw by calling
        # <code>updateGL()</code>.
        elif event.buttons() & QtCore.Qt.LeftButton:
            self.modelview_matrix *= rotation_matrix(-sqrt(dx ** 2 + dy ** 2), (dy, dx, 0.0), degrees=True)
            self.updateGL()

        # If the right mouse button is down, we scale the modelview matrix
        # dependent on the vertical mouse movement.  Again, a screen redraw is
        # triggered by calling <code>updateGL()</code>.
        elif event.buttons() & QtCore.Qt.RightButton:
            self.modelview_matrix *= scale_matrix(exp(-0.01 * dy))
            self.updateGL()

        # Finally, the mouse position is stored for processing of the following
        # mouse event.
        self.lastPos = QtCore.QPoint(event.pos())

    # <h4>OpenGL interaction</h4>

    # Whenever the canvas is resized (also on creation), we set the viewport
    # (which is actually a <i>glitter</i> <code>Context</code> property) to the
    # full window and change the projection matrix such that it encompasses the
    # whole range from -1 to +1:
    def resizeGL(self, w, h):
        self.viewport = (0, 0, w, h)
        self.projection_matrix = scale_matrix(h / float(w) if w > h else 1.0, w / float(h) if h > w else 1.0, 0.4)

    # The <code>paintGL()</code> method is called by Qt whenever the canvas
    # needs to be redrawn:
    def paintGL(self):
        # First, we clear the canvas using <code>Context.clear()</code>.
        self.clear()

        # Then, if a mesh pipeline has been loaded, we draw it with the
        # <code>modelview_matrix</code> uniform set. The pipeline binds the
        # vertex array and the shader, sets the depth test we requested
        # earlier, draws the vertex array, and resets all modified state:
        if self.mesh is not None:
            self.mesh.draw_with(modelview_matrix=self.projection_matrix * self.modelview_matrix)

    # Finally, when the user requests resetting the view via the menu system,
    # we create a clean modelview matrix and update the screen:
    def resetView(self):
        self.modelview_matrix = identity_matrix()
        self.updateGL()

# <h3>Qt Main Window</h3>

# The OpenGL widget will be displayed within a <code>QMainWindow</code> that
# provides menus, keyboard shortcuts, and many other amenities:
class MainWindow(QtGui.QMainWindow):
    # <h4>Initialization</h4>
    def __init__(self, parent=None):
        super(MainWindow, self).__init__(parent)

        # First, we create an instance of the <code>Canvas</code> class we
        # defined previously and make it the main widget in the window:
        self.canvas = Canvas()
        self.setCentralWidget(self.canvas)

        # Then, we create menus and keyboard shortcuts for opening meshes,
        # resetting the view, and exiting the program:
        self.fileMenu = self.menuBar().addMenu("&File")
        self.fileOpenMeshAction = QtGui.QAction("&Open Mesh\u2026", self)
        self.fileOpenMeshAction.setShortcut(QtGui.QKeySequence(QtCore.Qt.Key_O))
        self.fileOpenMeshAction.triggered.connect(self.fileOpenMesh)
        self.fileMenu.addAction(self.fileOpenMeshAction)
        self.fileMenu.addSeparator()
        self.fileQuitAction = QtGui.QAction("&Quit", self)
        self.fileQuitAction.setShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Escape))
        self.fileQuitAction.triggered.connect(self.close)
        self.fileMenu.addAction(self.fileQuitAction)

        self.viewMenu = self.menuBar().addMenu("&View")
        self.viewResetAction = QtGui.QAction("&Reset", self)
        self.viewResetAction.setShortcut(QtGui.QKeySequence(QtCore.Qt.Key_R))
        self.viewResetAction.triggered.connect(self.canvas.resetView)
        self.viewMenu.addAction(self.viewResetAction)

    # <h4>File loading</h4>
    # When the file open action is triggered via the menu or a keyboard
    # shortcut, we display a dialog for selecting a file. If a file is
    # selected, we tell the canvas to load a mesh from it.
    def fileOpenMesh(self, filename=None):
        if filename is None:
            dialog = QtGui.QFileDialog(self, "Open Mesh")
            dialog.setNameFilters(["%s files (*.%s)" % (x.upper(), x) for x in load_mesh.supported_formats])
            dialog.setViewMode(QtGui.QFileDialog.Detail)
            dialog.setAcceptMode(QtGui.QFileDialog.AcceptOpen)
            dialog.setFileMode(QtGui.QFileDialog.ExistingFile)
            if dialog.exec_():
                filename = dialog.selectedFiles()[0]
        if filename is not None:
            self.canvas.loadMesh(filename)

# <h2>Main section</h2>
# Finally, if this program is being run from the command line, we create a
# <code>QApplication</code>, show a <code>MainWindow</code> instance and run
# the Qt main loop.
if __name__ == "__main__":
    import sys
    app = QtGui.QApplication(sys.argv)
    mainWindow = MainWindow()
    mainWindow.show()
    sys.exit(app.exec_())

# When the main window is closed, the application will exit cleanly.




from glitter import Sampler, EnumConstant

def check_property(sampler, name):
    value = getattr(sampler, name)
    if isinstance(value, EnumConstant):
        valid_values = list(value._enum._reverse_dict.values())
        for value in valid_values:
            setattr(sampler, name, value)
            assert getattr(sampler, name) == value, "property %s is broken" % name
    else:
        setattr(sampler, name, value)
        assert getattr(sampler, name) == value, "property %s is broken" % name

def test_property_generator():
    sampler = Sampler(0)
    properties = [x for x in dir(sampler) if x != "context" and not x.startswith("_") and type(getattr(Sampler, x)) == property]

    for p in properties:
        yield check_property, sampler, p




"""Internal helper classes and functions.

@author: Stephan Wenger
@date: 2012-02-29
"""

try:
    from contextlib import nested
except ImportError:
    pass

from glitter.utils import constants
from glitter.utils.constants import *
from glitter.utils.dtypes import *
from glitter.utils.enum import *
from glitter.utils.errors import *
from glitter.utils.proxy import *
from glitter.utils.shaderdtypes import *
from glitter.utils.objects import *




"""Qt context creation and management.

@author: Stephan Wenger
@date: 2012-02-29
"""

try:
    from PySide import QtOpenGL
except ImportError:
    from PyQt4 import QtOpenGL

from glitter.contexts.context import Context

class QtContextWrapper(Context):
    def __init__(self, context):
        super(QtContextWrapper, self).__init__(self)
        if isinstance(context, QtOpenGL.QGLContext):
            self._qt_context = context
        else:
            self._qt_context = context.context()

    def _bind(self):
        self._qt_context.makeCurrent()

class QtWidget(QtOpenGL.QGLWidget, Context):
    def __init__(self, *args, **kwargs):
        QtOpenGL.QGLWidget.__init__(self, *args, **kwargs)
        with self:
            Context.__init__(self)

    def _bind(self):
        self.makeCurrent()




"""Shader classes.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils import ManagedObject, ShaderCompileError

class Shader(ManagedObject):
    _generate_id = _gl.glCreateShader
    _delete_id = _gl.glDeleteShader
    _db = "shaders"

    def __init__(self, source=None, compile=None, context=None):
        if any(x is NotImplemented for x in (self._type,)):
            raise TypeError("%s is abstract" % self.__class__.__name__)
        super(Shader, self).__init__(context=context)
        if source:
            self.source = source
        if compile is None:
            compile = bool(source)
        if compile:
            self.compile()

    def compile(self):
        with self._context:
            _gl.glCompileShader(self._id)
        
            _compile_status = _gl.GLint()
            _gl.glGetShaderiv(self._id, _gl.GL_COMPILE_STATUS, _gl.pointer(_compile_status))
            if _compile_status.value != _gl.GL_TRUE:
                raise ShaderCompileError(self._log)

            return self._log or None

    @property
    def _log(self):
        with self._context:
            _info_log_length = _gl.GLint()
            _gl.glGetShaderiv(self._id, _gl.GL_INFO_LOG_LENGTH, _gl.pointer(_info_log_length))
            _info_log = _gl.create_string_buffer(_info_log_length.value)
            _gl.glGetShaderInfoLog(self._id, _info_log_length, _gl.POINTER(_gl.GLint)(), _info_log)
            return _info_log.value

    @property
    def source(self):
        with self._context:
            _shader_source_length = _gl.GLint()
            _gl.glGetShaderiv(self._id, _gl.GL_SHADER_SOURCE_LENGTH, _gl.pointer(_shader_source_length))
            _source = _gl.create_string_buffer(_shader_source_length.value)
            _gl.glGetShaderSource(self._id, _shader_source_length, _gl.POINTER(_gl.GLint)(), _source)
            return _source.value

    @source.setter
    def source(self, source):
        with self._context:
            _gl.glShaderSource(self._id, 1, _gl.pointer(_gl.c_char_p(source.encode("ascii"))), _gl.POINTER(_gl.GLint)())

class VertexShader(Shader):
    """Vertes shader.

    @todo: wrap C{glGetShaderPrecisionFormat}.
    """

    _type = _gl.GL_VERTEX_SHADER

class TesselationControlShader(Shader):
    _type = _gl.GL_TESS_CONTROL_SHADER

class TesselationEvaluationShader(Shader):
    _type = _gl.GL_TESS_EVALUATION_SHADER

class GeometryShader(Shader):
    _type = _gl.GL_GEOMETRY_SHADER

class FragmentShader(Shader):
    """Fragment shader.
    
    @todo: wrap C{glGetShaderPrecisionFormat}.
    """
    
    _type = _gl.GL_FRAGMENT_SHADER

__all__ = [
    "Shader",
    "VertexShader",
    "TesselationControlShader",
    "TesselationEvaluationShader",
    "GeometryShader",
    "FragmentShader",
]




#!/usr/bin/env python

#! This file is a literate Python program. You can compile the documentation
#! using mylit (http://pypi.python.org/pypi/mylit/).
## title = "glitter Example: Volume Renderer"
## stylesheet = "pygments_style.css"

# <h1><i>glitter</i> Example: Volume Renderer</h1>

# <h2>Summary</h2>

# This program will create an invisible OpenGL context to render a volume given
# as an <a href="www.hdfgroup.org/HDF5/">HDF5</a> data file to images that are
# written to files.

# We will use a very simple algorithm for volume rendering. First, the back
# face of the bounding box of the volume will be rendered into a texture using
# front face culling. The texture coordinates (or positions) of the fragments
# are stored in the texture. Then, in a second pass, the front face of the
# bounding box is rendered using back face culling. The shader reads the
# texture coordinates of the previous step from the texture, so that it now
# knows about both the viewing ray entry point to the bounding box (from the
# current texture coordinates) as well as the ray exit point (from the texture
# coordinates stored in the texture). The shader then uniformly samples a 3D
# texture along the line between these points and sums up the intensities. This
# is a valid image formation model for purely emissive volumes. To simulate an
# absorbing volume illuminated uniformly from the back, an exponential
# transform can optionally be applied to the final pixel intensities.

# <h2>Front matter</h2>

# <h3>Module docstring</h3>

# The module docstring is used as a description of this example in the
# generated documentation:
"""Volume renderer that creates a set of image files from different perspectives.

@author: Stephan Wenger
@date: 2012-03-07
"""

# <h3>Imports</h3>

# We will use <a
# href="http://docs.python.org/library/itertools.html">itertools</a> to create
# the vertices of a cube:
import itertools

# <a href="http://numpy.scipy.org/">numpy</a> is used for array operations:
import numpy

# We can usually import classes and functions contained in <i>glitter</i>
# submodules directly from glitter:
from glitter import Framebuffer, ShaderProgram, RectangleTexture, VertexArray, float32, State, Texture3D

# <h2>Shaders</h2>

# The vertex shader transforms the position as usual; additionally, it stores
# the unmodified vertex coordinates for use in the fragment shader:
vertex_code = """
#version 400 core

layout(location=0) in vec4 in_position;
uniform mat4 modelview_matrix;
out vec4 ex_front;

void main() {
	ex_front = in_position;
	gl_Position = modelview_matrix * 2.0 * (in_position - 0.5) * vec4(1.0, 1.0, 0.5, 1.0);
}
"""

# We need to different fragment shaders: one for the back faces, and one for
# the front faces. The shader for the back faces simply stores the
# untransformed position in the color output:
back_fragment_code = """
#version 400 core
#extension GL_ARB_texture_rectangle : enable

in vec4 ex_front;
layout(location=0) out vec4 out_color;

void main() {
    out_color = ex_front;
}
"""

# The fragment shader for the front faces performs the actual integration along
# the viewing ray:
front_fragment_code = """
#version 400 core
#extension GL_ARB_texture_rectangle : enable

#define STEP_SIZE 0.001
#define MAX_LENGTH sqrt(3.0)

in vec4 ex_front;
uniform sampler2DRect back;
uniform sampler3D volume;
uniform float intensity_scale;
uniform bool absorption;
layout(location=0) out vec4 out_color;

void main() {
	vec3 p0 = texture(back, gl_FragCoord.st).xyz;
	vec3 p1 = ex_front.xyz;
	vec4 intensity = vec4(0.0);

	vec3 d = p1 - p0;
	float l = length(d);
	if (l < MAX_LENGTH) {
		vec3 s = d * STEP_SIZE / l;
		vec3 pos = p0 + 0.5 * s;

		for (float f = 0.5 * STEP_SIZE; f < l; f += STEP_SIZE) {
			intensity += texture(volume, pos);
			pos += s;
		}
	}
	
    out_color = absorption ? exp(-intensity_scale * intensity) : intensity_scale * intensity;
}
"""

# <h2>Vertex arrays</h2>

# Here we define the faces of a cube that will serve as a bounding box to the
# volume:
cube_vertices = tuple(itertools.product((0.0, 1.0), (0.0, 1.0), (0.0, 1.0)))
cube_indices = ((0, 1, 3), (0, 2, 6), (0, 3, 2), (0, 6, 4), (3, 1, 5), (3, 5, 7),
        (4, 1, 0), (4, 5, 1), (5, 4, 6), (5, 6, 7), (7, 2, 3), (7, 6, 2))

# <h2>Volume renderer</h2>

# The <code>VolumeRenderer</code> class performs the actual volume rendering:
class VolumeRenderer(object):
    # Some default values for instance variables can be declared as class
    # variables (these values are all immutable, so no problem here):
    modelview_matrix = tuple(map(tuple, numpy.eye(4)))
    intensity_scale = 1.0
    absorption = False

    # <h3>Initialization</h3>
    def __init__(self, volume, **kwargs):
        assert isinstance(volume, Texture3D), "volume must be a 3D texture"

        # We need a vertex array to hold the faces of the cube:
        self.vao = VertexArray(cube_vertices, elements=cube_indices)

        # For rendering the back faces, we create a shader program and a
        # framebuffer with an attached texture:
        self.back_shader = ShaderProgram(vertex=vertex_code, fragment=back_fragment_code)
        self.back_fbo = Framebuffer(RectangleTexture(shape=volume.shape[:2] + (4,), dtype=float32))

        # For the front faces, we do the same. Additionally, the shader
        # receives the texture containing the back face coordinates in a
        # uniform variable <code>back</code>, and the 3D texture containing the
        # volumetric data in a uniform variable <code>volume</code>.
        self.front_shader = ShaderProgram(vertex=vertex_code, fragment=front_fragment_code, back=self.back_fbo[0], volume=volume)
        self.front_fbo = Framebuffer(RectangleTexture(shape=volume.shape[:2] + (4,), dtype=float32))

        # All other parameters are simply set as attributes (this might be
        # <code>modelview_matrix</code>, <code>intensity_scale</code>, or
        # <code>absorption</code>).
        for key, value in list(kwargs.items()):
            setattr(self, key, value)

    # <h3>Rendering</h3>
    def render(self):
        # We first draw the back faces, storing their coordinates into a
        # texture. This means we have to clear and bind the back framebuffer,
        # activate face culling with cull face mode "FRONT", bind the back face
        # shader with the current modelview matrix set, and render the vertex
        # array for the cube:
        self.back_fbo.clear()
        with State(cull_face=True, cull_face_mode="FRONT"):
            with self.back_shader(modelview_matrix=self.modelview_matrix):
                with self.back_fbo:
                    self.vao.draw()

        # We then draw the front faces, accumulating the intensity between
        # front and back face. The setup is very similar to the one stated
        # before, except that the additional uniform variables
        # <code>intensity_scale</code> and <code>absorption</code> are passed
        # to the shader:
        self.front_fbo.clear()
        with State(cull_face=True, cull_face_mode="BACK"):
            with self.front_shader(modelview_matrix=self.modelview_matrix, intensity_scale=self.intensity_scale, absorption=self.absorption):
                with self.front_fbo:
                    self.vao.draw()

        # Finally, we simply return the texture into which we just rendered.
        # The client code might want to display this directly, or to download
        # the data into a numpy array.
        return self.front_fbo[0]

# <h2>Initialization and main loop</h2>

# Finally, if this program is being run from the command line, we set up all
# the previously mentioned objects and start rendering. The program will expect
# two command line parameters: the name of an HDF5 input file, and a
# <code>printf</code>-style format string for the image filenames.
if __name__ == "__main__":
    # We need to read a volume filename from <code>sys.argv</code>, so import
    # <code>sys</code>.
    import sys

    # We assume the volume is stored in a <a
    # href="http://www.hdfgroup.org/HDF5/">HDF5</a> file, so import <a
    # href="h5py.alfven.org"><code>h5py</code></a>.
    import h5py

    # For saving the resulting images, we use <a
    # href="http://scipy.org/">SciPy</a>'s <code>imread</code> function:
    from scipy.misc import imsave

    # We open the HDF5 file specified on the command line for reading:
    with h5py.File(sys.argv[1]) as f:
        # The volumetric data is read from the corresponding dataset in the
        # HDF5 file. Note that the name of the dataset is mere convention.
        volume = Texture3D(f["data"])
        # The data can be either emission densities only or absorption
        # densities only.  If the dataset defines an attribute called
        # "apsorption", we read it; otherwise we assume it's an emission only
        # dataset.
        absorption = f["data"].attrs.get("absorption", False)

    # Now we generate a <code>VolumeRenderer</code> instance for the volumetric
    # data that has just been read.
    renderer = VolumeRenderer(volume, absorption=absorption, intensity_scale=0.1, modelview_matrix=numpy.eye(4))

    # Finally, we iterate over a number of different viewing angles, manipulate
    # the modelview matrix accordingly, render the volume, and save it to an
    # image file:
    maxval = None
    for idx, angle in enumerate(numpy.mgrid[0:2*numpy.pi:100j]):
        renderer.modelview_matrix[::2, ::2] = numpy.array(((numpy.cos(angle), numpy.sin(angle)), (-numpy.sin(angle), numpy.cos(angle))))
        # The <code>render()</code> function returns a <code>Texture</code>
        # object. We can access the texture data via the <code>data</code>
        # attribute. We mirror the image vertically and chop of the alpha
        # channel before storing the image in a file:
        image = renderer.render().data[::-1, :, :3]
        if maxval is None:
            maxval = image.max()
        imsave(sys.argv[2] % idx, image / maxval)

    # Note how we did not need to create an OpenGL context explicitly. The
    # first object that needs an OpenGL context (in this case the
    # <code>VertexArray</code> in the <code>VolumeRenderer</code>) will create
    # an invisible context as soon as it is initialized. This is very
    # convenient; on the other hand, it means that if you create a context
    # manually (e.g., because you need a visible window), you have to
    # initialize it before any other objects. Otherwise, these objects may end
    # up in the wrong context, and sharing data across context is a problem of
    # its own.




"""Descriptors for L{ShaderProgram} uniforms.

End users should typically not need to use this module directly.

@bug: Structs and arrays of structs are currently unimplemented.

@author: Stephan Wenger
@date: 2012-02-29
"""

from collections import OrderedDict as _odict
import numpy as _np

import glitter.raw as _gl

class BaseUniform(object):
    pass

class Uniform(BaseUniform):
    def __init__(self, name, location, dtype, size, parent):
        self.name = name
        self.location = location
        self.dtype = dtype
        self.size = size
        self.parent = parent
        self.textures = []

    def __str__(self):
        return "uniform %s %s[%d];" % (self.dtype, self.name, self.size)

    def __get__(self, obj, cls=None, get_texture_unit_instead_of_object=False): # TODO return a writable handle to arrays
        if self.dtype.is_texture() and not get_texture_unit_instead_of_object:
            return self.textures[0] if len(self.textures) == 1 else self.textures
        else:
            with obj._context:
                if self.size == 1:
                    data = self.dtype.get_value(obj, self.location)
                    return data.item() if len(data) == 1 else data
                else:
                    data = [self.dtype.get_value(obj, _gl.glGetUniformLocation(obj._id, ("%s[%d]" % (self.name, i)).encode("ascii"))) for i in range(self.size)]
                    return _np.concatenate([x.squeeze()[None] for x in data])

    def __set__(self, obj, value):
        if self.dtype.is_texture():
            if self.parent._context.current_program == self.parent:
                self._on_release()
            self.textures = value if hasattr(value, "__iter__") and not isinstance(value, str) else [value]
            if self.parent._context.current_program == self.parent:
                self._on_bind()
        else:
            self.dtype.set_value(obj, self.location, value, self.size)

    def _on_bind(self):
        if self.dtype.is_texture():
            units = [self.parent._context.texture_units.bind(x) for x in self.textures]
            if units:
                self.dtype.set_value(self.parent, self.location, units, self.size)

    def _on_release(self):
        if self.dtype.is_texture():
            for x in self.textures:
                self.parent._context.texture_units.release(x)

class UniformStruct(_odict, BaseUniform):
    def __init__(self, name, parent):
        super(UniformStruct, self).__init__()
        self.name = name
        self.parent = parent

    def __str__(self):
        return "uniform struct { %s } %s;" % (" ".join(str(value) for value in list(self.values())), self.name)

    def __get__(self, obj, cls=None):
        """@todo: Implement this."""
        raise NotImplementedError

    def __set__(self, obj, value):
        """@todo: Implement this."""
        raise NotImplementedError

    def _on_bind(self):
        """@todo: Implement this: bind textures."""
        pass

    def _on_release(self):
        """@todo: Implement this: restore old texture bindings."""
        pass

class UniformStructArray(_odict, BaseUniform):
    def __init__(self, name, parent):
        super(UniformStructArray, self).__init__()
        self.name = name
        self.parent = parent

    @property
    def size(self):
        return max(index for index, field in list(self.keys())) + 1

    def __str__(self):
        unique_values = _odict((field, value) for ((index, field), value) in list(self.items()))
        return "uniform struct { %s } %s[%d];" % (" ".join(str(value) for value in list(unique_values.values())), self.name, self.size)

    def __get__(self, obj, cls=None):
        """@todo: Implement this."""
        raise NotImplementedError

    def __set__(self, obj, value):
        """@todo: Implement this."""
        raise NotImplementedError

    def _on_bind(self):
        """@todo: Implement this: bind textures."""
        pass

    def _on_release(self):
        """@todo: Implement this: restore old texture bindings."""
        pass




"""Base classes for descriptors and descriptor owners.

@bug: L{Proxy} is currently unaware of its context.
@author: Stephan Wenger
@date: 2012-02-29
"""

from inspect import getmembers as _getmembers
import types as _types
import numpy as _np

import glitter.raw as _gl
from glitter.utils.dtypes import coerce_array, int32
from glitter.utils.objects import with_obj
import collections

class Proxy(object):
    def __init__(self, getter=None, get_args=(), setter=None, set_args=(), dtype=None, shape=(), enum=None, name=None):
        self._getter = getter
        self._get_args = get_args
        self._setter = setter
        self._set_args = set_args
        self._dtype = dtype if dtype else int32
        self._shape = None if shape is None else tuple(shape) if hasattr(shape, "__iter__") and not isinstance(shape, str) else (shape,)
        self._enum = enum
        self._name = name

    def __get__(self, obj, cls):
        _value = _np.empty(self._shape, dtype=self._dtype.as_numpy())
        args = list(self._get_args) + [_gl.cast(_value.ctypes, self._getter.argtypes[-1])]
        with obj:
            self._getter(*args)
        value = _value.item() if _value.shape is () else _value
        if self._enum is not None:
            if hasattr(value, "__iter__") and not isinstance(value, str):
                value = [self._enum[x] for x in value]
            else:
                value = self._enum[value]
        return value

    def __set__(self, obj, value):
        if self._setter is None:
            raise AttributeError("can't set attribute")
        if self._enum is not None:
            if hasattr(value, "__iter__") and not isinstance(value, str):
                _value = [self._enum(x)._value for x in value]
            else:
                _value = self._enum(value)._value
        else:
            _value = value
        _value = coerce_array(_value, dtype=self._dtype)
        if len(self._set_args) + len(_value) == len(self._setter.argtypes):
            args = list(self._set_args) + (list(_value) if _value.ndim == 1 else [x.ctypes for x in _value])
        elif len(self._set_args) + 1 == len(self._setter.argtypes):
            args = list(self._set_args) + [_value.ctypes]
        else:
            raise RuntimeError("no valid setter invocation found")
        with obj:
            self._setter(*args)

    def __repr__(self):
        if self._name:
            return "proxy for %s" % self._name
        elif self._getter and self._setter:
            return "proxy for %s(%s) / %s(%s)" % (
                    self._getter.__name__, ", ".join(map(str, self._get_args)),
                    self._setter.__name__, ", ".join(map(str, self._set_args)))
        elif self._getter:
            return "proxy for %s(%s)" % (self._getter.__name__, ", ".join(map(str, self._get_args)))
        elif self._setter:
            return "proxy for %s(%s)" % (self._setter.__name__, ", ".join(map(str, self._set_args)))
        else:
            return "proxy object"

class ListProxy(object):
    def __init__(self, lst, insert_callback=None, delete_callback=None):
        self._lst = lst
        self._insert_callback = insert_callback
        self._delete_callback = delete_callback

    def append(self, x):
        self._lst.append(x)
        self._insert_callback(x)

    def extend(self, xs):
        for x in xs:
            self.append(x)

    def remove(self, x):
        self._lst.remove(x)
        self._delete_callback(x)

    def __iadd__(self, xs):
        self.extend(xs)

    def __getitem__(self, key):
        return self._lst[key]

    def __len__(self):
        return len(self._lst)

class InstanceDescriptorMixin(object):
    """Mixin to enable runtime-added descriptors."""

    def __getattribute__(self, name):
        attr = super(InstanceDescriptorMixin, self).__getattribute__(name)
        if hasattr(attr, "__get__") and not isinstance(attr, collections.Callable):
            return attr.__get__(self, self.__class__)
        else:
            return attr

    def __setattr__(self, name, value):
        try:
            attr = super(InstanceDescriptorMixin, self).__getattribute__(name)
            return attr.__set__(self, value)
        except AttributeError:
            return super(InstanceDescriptorMixin, self).__setattr__(name, value)

class PropertyProxy(object):
    def __init__(self, obj, name):
        self._obj = obj
        self._name = name

    def __get__(self, obj, cls):
        return getattr(self._obj, self._name)

    def __set__(self, obj, value):
        setattr(self._obj, self._name, value)

    def __delete__(self, obj):
        delattr(self._obj, self._name)

class ItemProxy(object):
    def __init__(self, obj, idx):
        self._obj = obj
        self._idx = idx

    def __get__(self, obj, cls):
        return self._obj[self._idx]

    def __set__(self, obj, value):
        self._obj[self._idx] = value

    def __delete__(self, obj):
        del self._obj[self._idx]

def add_proxies(parent, obj):
    """Add proxies for methods and properties of C{obj} to C{parent}."""

    # add functions and methods
    for key, value in _getmembers(obj):
        if key.startswith("_"):
            continue
        if isinstance(value, (_types.FunctionType, _types.MethodType)):
            setattr(parent, key, with_obj(parent, value))

    # add properties and other descriptors (can be accessed via the class only)
    for key, value in _getmembers(type(obj)):
        if key.startswith("_"):
            continue
        if hasattr(value, "__get__") and not isinstance(value, collections.Callable): # exclude member functions
            setattr(parent, key, PropertyProxy(obj, key))

    # add instance descriptors if applicable
    if isinstance(obj, InstanceDescriptorMixin):
        for key, value in list(obj.__dict__.items()):
            if key.startswith("_"):
                continue
            if hasattr(value, "__get__"):
                setattr(parent, key, PropertyProxy(obj, key))
        
__all__ = ["Proxy", "ListProxy", "InstanceDescriptorMixin", "PropertyProxy", "ItemProxy", "add_proxies"]




"""Renderbuffer object class.

@bug: Renderbuffers are currently unimplemented.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils import ManagedObject, BindableObject

class Renderbuffer(ManagedObject, BindableObject):
    """Renderbuffer class.
    
    @todo: Use C{glFramebufferRenderbuffer} instead of L{_binding}; returns C{GLsync}!
    @todo: Wrap C{glRenderbufferStorage} and C{glRenderbufferStorageMultisample}.
    """

    _generate_id = _gl.glGenRenderbuffers
    _delete_id = _gl.glDeleteRenderbuffers
    _db = "renderbuffers"

    def __init__(self, context=None):
        super(Renderbuffer, self).__init__(context=context)

__all__ = ["Renderbuffer"]




#!/usr/bin/env python

#! This file is a literate Python program. You can compile the documentation
#! using mylit (http://pypi.python.org/pypi/mylit/).
## title = "glitter Example: Voxelization"
## stylesheet = "pygments_style.css"

# <h1><i>glitter</i> Example: Voxelization</h1>

# <h2>Summary</h2>

# This program will create an invisible OpenGL context to voxelize a mesh given
# as an <a href="www.hdfgroup.org/HDF5/">HDF5</a> data file. The resulting
# volume is written to a file.

#! TODO describe voxelization algorithm
#! TODO include spreadbits.py so the data can be viewed in the volume viewer

# <h2>Front matter</h2>

# <h3>Module docstring</h3>

# The module docstring is used as a description of this example in the
# generated documentation:
"""Voxelize a mesh.

@author: Stephan Wenger
@date: 2012-02-29
"""

# We can usually import classes and functions contained in <i>glitter</i>
# submodules directly from glitter:
from glitter import VertexArray, TextureArray2D, uint32, Framebuffer, State, ShaderProgram

# <h2>Shaders</h2>

# The vertex shader transforms the position as usual.
vertex_code = """
#version 400 core

layout(location=0) in vec4 in_position;
out vec4 ex_position;

void main() {
    gl_Position = ex_position = in_position;
}
"""

# The fragment code will perform the actual work. It will quantize each
# fragment into a depth layer, compute a bitmask corresponding to that depth,
# and write the bit mask into the color attachments. The number of color
# attachments is only known at runtime, which is why a <code>%d</code>
# placeholder will have to be filled in just before the shader is compiled.
fragment_code = """
#version 400 core

in vec4 ex_position;
uniform bool solid;
layout(location=0) out uvec4 fragmentColor[%d];

uvec4 voxelize(uint z, uint offset, uint fill, uint base) {
    if (z >= (offset + 1u) * 128u) return uvec4(0u);
    if (z < offset * 128u) return uvec4(fill);
    z -= offset * 128u;
    return uvec4(
        z < 128u ? (z >= 96u ? base << (z - 96u) : fill) : 0u,
        z <  96u ? (z >= 64u ? base << (z - 64u) : fill) : 0u,
        z <  64u ? (z >= 32u ? base << (z - 32u) : fill) : 0u,
        z <  32u ?             base << (z -  0u)         : 0u
    );
}

void main() { // if solid, switch on all bits >= z; else, switch on bit == z
    uint z = uint(round(128.0 * float(fragmentColor.length()) * 0.5 * (ex_position.z + 1.0)));
    for (int i = 0; i < fragmentColor.length(); ++i)
        fragmentColor[i] = voxelize(z, i, solid ? 0xffffffffu : 0u, solid ? 0xffffffffu : 1u);
}
"""

# <h2>Voxelization</h2>

# The function <code>voxelize()</code> performs solid or boundary voxelization
# of a mesh into a volume of given size.
def voxelize(mesh, size, solid=True):
    """Voxelize a mesh into a cube of length C{size}.

    @param mesh: The mesh to voxelize.
    @type mesh: L{VertexArray}
    @param size: The length of the resulting cube. Should be a multiple of 128.
    @type size: C{int}
    @param solid: Whether to perform solid voxelization instead of boundary voxelization.
    @type solid: C{bool}

    @attention: The result of boundary voxelization is likely to have holes. To
    obtain a watertight volume, you can voxelize the volume from different
    directions and combine the results, or use the gradient of a solid volume.
    """

    # An array of 2D integer textures will hold the output volume encoded in
    # its bits:
    volume = TextureArray2D(shape=(size // 128, size, size, 4), dtype=uint32)

    # We create a framebuffer with all layers of the volume attached and
    # directly bind it:
    with Framebuffer(*[volume[i] for i in range(len(volume))]) as fbo:
        fbo.clear()
        # Depending on the voxelization mode, we set the logic op mode. We then
        # create and activate a shader program and draw the mesh.
        with State(logic_op_mode="XOR" if solid else "OR", color_logic_op=True):
            with ShaderProgram(vertex=vertex_code, fragment=fragment_code % len(volume), solid=solid):
                mesh.draw()

    # Finally, the texture array is returned. The client code can decide to
    # download the data or to process it further on the GPU.
    return volume

# <h2>Initialization and main loop</h2>

# Finally, if this program is being run from the command line, we set up all
# the previously mentioned objects and start rendering. The program will expect
# two command line parameters: the name of an HDF5 input file, and the name of
# an HDF5 output file. An optional third parameter specifies the size of the
# resulting volume.
if __name__ == "__main__":
    # We need to read a mesh filename from <code>sys.argv</code>, so import
    # <code>sys</code>.
    import sys

    # We assume the mesh is stored in a <a
    # href="http://www.hdfgroup.org/HDF5/">HDF5</a> file, so import <a
    # href="h5py.alfven.org"><code>h5py</code></a>.
    import h5py

    # We open the HDF5 file specified on the command line for reading, extract
    # the vertices and indices, and store them in a vertex array:
    with h5py.File(sys.argv[1]) as f:
        mesh = VertexArray(f["vertices"], elements=f["indices"])

    # Now the voxelization is performed and the result is returned as an array
    # of 2D integer textures.
    volume = voxelize(mesh, int(sys.argv[3]) if len(sys.argv) > 3 else 128)

    # Finally, we open the output file, download the data, and store it in a
    # dataset. The dataset can typically be compressed a lot, which is why we
    # enable LZF compression. However, this may take a long time, so you can
    # disable it if you are not short on disk space.
    with h5py.File(sys.argv[2], "w") as f:
        f.create_dataset("data", data=volume.data, compression="lzf")

    # Note how we did not need to create an OpenGL context explicitly. The
    # first object that needs an OpenGL context (in this case the
    # <code>VertexArray</code>) will create an invisible context as soon as it
    # is initialized. This is very convenient; on the other hand, it means that
    # if you create a context manually (e.g., because you need a visible
    # window), you have to initialize it before any other objects. Otherwise,
    # these objects may end up in the wrong context, and sharing data across
    # context is a problem of its own.




#!/usr/bin/env python

"""Poisson integration on OpenGL textures using OpenCL.

@todo: Add literate programming comments.

@author: Stephan Wenger
@date: 2012-03-20
"""

from numpy import float32, zeros_like
import pyopencl as cl
from pyopencl.tools import get_gl_sharing_context_properties

from glitter import Texture2D
from glitter.raw.gl import GL_TEXTURE_2D

kernel_code = """
__kernel void laplace(__write_only image2d_t dst, __read_only image2d_t src, __read_only image2d_t lapl_mask, float a) {
    const sampler_t sampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP_TO_EDGE | CLK_FILTER_NEAREST;
    int2 coords = (int2)(get_global_id(0), get_global_id(1));

    float4 lapl_val = read_imagef(lapl_mask, sampler, coords);
    if (lapl_val.w < 0.5) {
        write_imagef(dst, coords,
            (1 - a) * read_imagef(src, sampler, coords) + 0.25 * a * (
              - lapl_val
              + read_imagef(src, sampler, coords + (int2)(1, 0))
              + read_imagef(src, sampler, coords - (int2)(1, 0))
              + read_imagef(src, sampler, coords + (int2)(0, 1))
              + read_imagef(src, sampler, coords - (int2)(0, 1))
            )
        );
     }
}
"""

class PoissonIntegrator(object):
    """Solve  Laplace(img) == lapl_mask  wherever lapl_mask.w is smaller than 0.5."""

    use_gl_texture_as_tmp = True # TODO when False, result looks strange; why?

    def __init__(self, gl_image, gl_lapl_mask):
        platform = cl.get_platforms()[0]
        self.ctx = cl.Context(properties=[(cl.context_properties.PLATFORM, platform)] + get_gl_sharing_context_properties(), devices=None)
        self.queue = cl.CommandQueue(self.ctx)
        self.program = cl.Program(self.ctx, kernel_code).build()

        self.shape = gl_image.shape[1], gl_image.shape[0]
        self.cl_image = cl.GLTexture(self.ctx, cl.mem_flags.READ_WRITE, GL_TEXTURE_2D, 0, gl_image._id, 2)
        self.cl_lapl_mask = cl.GLTexture(self.ctx, cl.mem_flags.READ_ONLY, GL_TEXTURE_2D, 0, gl_lapl_mask._id, 2)

        if self.use_gl_texture_as_tmp:
            self.gl_tmp = Texture2D(gl_image.data)
            self.cl_tmp = cl.GLTexture(self.ctx, cl.mem_flags.READ_WRITE, GL_TEXTURE_2D, 0, self.gl_tmp._id, 2)
        else:
            self.cl_tmp = cl.Image(self.ctx, cl.mem_flags.READ_WRITE, self.cl_image.get_image_info(cl.image_info.FORMAT), self.shape)

    def execute(self, a=1.0):
        if self.use_gl_texture_as_tmp:
            cl.enqueue_acquire_gl_objects(self.queue, [self.cl_image, self.cl_lapl_mask, self.cl_tmp])
        else:
            cl.enqueue_acquire_gl_objects(self.queue, [self.cl_image, self.cl_lapl_mask])
        self.program.laplace(self.queue, self.shape, None, self.cl_tmp, self.cl_image, self.cl_lapl_mask, float32(a))
        self.program.laplace(self.queue, self.shape, None, self.cl_image, self.cl_tmp, self.cl_lapl_mask, float32(a))
        if self.use_gl_texture_as_tmp:
            cl.enqueue_release_gl_objects(self.queue, [self.cl_image, self.cl_lapl_mask, self.cl_tmp])
        else:
            cl.enqueue_release_gl_objects(self.queue, [self.cl_image, self.cl_lapl_mask])

if __name__ == "__main__":
    import sys
    from matplotlib.pyplot import imread
    from scipy.misc import imsave

    image = imread(sys.argv[1])
    lapl_mask = zeros_like(image)
    lapl_mask[:, :, 3] = image[:, :, 3]
    lapl_mask[0, :, 3] = lapl_mask[-1, :, 3] = lapl_mask[:, 0, 3] = lapl_mask[:, -1, 3] = 1.0

    gl_image = Texture2D(image)
    gl_lapl_mask = Texture2D(lapl_mask)
    
    poisson_integrator = PoissonIntegrator(gl_image, gl_lapl_mask)
    for i in range(1024):
        poisson_integrator.execute(0.5)

    imsave(sys.argv[2], gl_image.data[:, :, :3])




#!/usr/bin/env python

#! This file is a literate Python program. You can compile the documentation
#! using mylit (http://pypi.python.org/pypi/mylit/).
## title = "glitter Example: OpenCL"
## stylesheet = "pygments_style.css"

# <h1><i>glitter</i> Example: OpenCL</h1>

# <h2>Summary</h2>

# This program will open a GLUT window and render an animated particle
# simulation. The particle simulation is computed in OpenCL, and the results
# are written directly to an OpenGL array buffer.

# <img src="opencl.png">

# <h2>Front matter</h2>

# <h3>Module docstring</h3>

# The module docstring is used as a description of this example in the
# generated documentation:
"""Basic OpenGL/OpenCL interoperability example.

@author: Stephan Wenger
@date: 2012-02-29
"""

# <h3>Imports</h3>

# We use numpy for creating the particle data:
import numpy

# <a href="http://mathema.tician.de/software/pyopencl">PyOpenCL</a> (<a
# href="http://documen.tician.de/pyopencl/">documentation</a>, <a
# href="http://pypi.python.org/pypi/pyopencl">PyPi</a>, <a
# href="http://sourceforge.net/projects/pyopencl/">download</a>) is a
# convenient Python wrapper for OpenCL:
import pyopencl as cl

# For creating an OpenCL context that shares data with OpenGL, we need
# <code>get_gl_sharing_context_properties</code>:
from pyopencl.tools import get_gl_sharing_context_properties

# We can usually import classes and functions contained in <i>glitter</i>
# submodules directly from glitter:
from glitter import ArrayBuffer, VertexArray, get_default_program

# Modules with external dependencies other than numpy, such as platform
# dependent parts like methods for the generation of an OpenGL context,
# however, have to be imported from their respective submodules:
from glitter.contexts.glut import GlutWindow, main_loop

# <h2>OpenCL</h2>

# <h3>OpenCL kernel</h3>

# The OpenCL kernel (adapted from a <a
# href="http://www.cmsoft.com.br/index.php?option=com_content&view=category&layout=blog&id=99&Itemid=150">CMSoft
# tutorial</a>) computes a single iteration of a simple Eulerian particle
# simulation. Kernel code can be loaded from a file or defined inline as a
# Python string:
kernel_code = """
__kernel void animate(__global float4* positions,
                      __global float4* colors,
                      __global float4* velocities,
                      __global float4* initial_positions,
                      __global float4* initial_velocities,
                      float dt) {
    unsigned int i = get_global_id(0);
    float4 p = positions[i];
    float4 v = velocities[i];
    float life = velocities[i].w;

    life -= dt;
    if (life <= 0.0) {
        p = initial_positions[i];
        v = initial_velocities[i];
        life = 1.0;    
    }

    v.z -= 9.8 * dt;
    p.xyz += v.xyz * dt;
    v.w = life;

    positions[i] = p;
    velocities[i] = v;
    colors[i].w = life;
}
"""

# <h3>OpenCL interaction</h3>

# The <code>CLCode</code> class will comprise all OpenCL interaction:
class CLCode(object):
    # <h4>Initialization</h4>
    # The constructor receives OpenGL buffers for positions and colors as well
    # as a numpy array of velocities and a timestep for the simulation:
    def __init__(self, gl_positions, gl_colors, velocities, dt=0.001):
        # First, we have to initialize the OpenCL context. That means we have
        # to get a list of available platforms and select one:
        platform = cl.get_platforms()[0]
        # Then, we can create a context. Passing
        # <code>get_gl_sharing_context_properties()</code> as a property
        # ensures that we share state with the active OpenGL context:
        self.ctx = cl.Context(properties=[(cl.context_properties.PLATFORM, platform)] +
                get_gl_sharing_context_properties(), devices=[platform.get_devices()[0]])
        # A command queue is necessary for serializing OpenCL commands:
        self.queue = cl.CommandQueue(self.ctx)
        # Finally, we can compile the kernel:
        self.program = cl.Program(self.ctx, kernel_code).build()

        # The constructor parameters are stored for later use:
        self.gl_positions = gl_positions
        self.gl_colors = gl_colors
        self.velocities = velocities
        # The <code>dt</code> value will later be passed to an OpenCL kernel as
        # a 32-bit float. We therefore wrap it in a numpy <code>float32</code>
        # object:
        self.dt = numpy.float32(dt)

        # Next, we generate OpenCL buffers. The positions and colors are
        # contained in OpenGL buffers, which we wrap in PyOpenCL's
        # <code>GLBuffer</code> class:
        self.cl_positions = cl.GLBuffer(self.ctx, cl.mem_flags.READ_WRITE, self.gl_positions._id)
        self.cl_colors = cl.GLBuffer(self.ctx, cl.mem_flags.READ_WRITE, self.gl_colors._id)
        # Note how we had to extract the <code>_id</code>s from the
        # <code>ArrayBuffer</code> objects. In pure <i>glitter</i> code, you
        # should never (have to) access this value; however for interaction
        # with other OpenGL-related libraries, this cannot always be avoided.

        # The velocities are given as a numpy array, which is simply uploaded
        # into a new OpenCL <code>Buffer</code> object along with the initial
        # values of the positions and colors:
        self.cl_velocities = cl.Buffer(self.ctx, cl.mem_flags.READ_ONLY |
                cl.mem_flags.COPY_HOST_PTR, hostbuf=velocities)
        self.cl_initial_positions = cl.Buffer(self.ctx, cl.mem_flags.READ_ONLY |
                cl.mem_flags.COPY_HOST_PTR, hostbuf=self.gl_positions.data)
        self.cl_initial_velocities = cl.Buffer(self.ctx, cl.mem_flags.READ_ONLY |
                cl.mem_flags.COPY_HOST_PTR, hostbuf=self.velocities)

    # <h4>Execution</h4>
    # The <code>execute</code> function executes the OpenCL kernel several
    # times in a row:
    def execute(self, sub_intervals):
        # First, we have to make sure that OpenGL is done using the buffer objects:
        cl.enqueue_acquire_gl_objects(self.queue, [self.cl_positions, self.cl_colors])
        # Now, we can safely call the kernel. Its arguments are buffer objects:
        args = (self.cl_positions, self.cl_colors, self.cl_velocities,
                self.cl_initial_positions, self.cl_initial_velocities, self.dt)
        # The kernel will be executed several times with a small step size.
        # This increases the accuracy with respect to a single step with a
        # large step size. However, it is not necessary to display all the
        # intermediate results.
        for i in range(0, sub_intervals):
            # In each step, the <code>animate</code> kernel function is called.
            # Its arguments are the queue object that schedules its execution,
            # the global and local block sizes, and any arguments that will be
            # passed to the actual kernel.
            self.program.animate(self.queue, [len(self.gl_positions)], None, *args)
        # Finally, we allow OpenGL to access the buffers again:
        cl.enqueue_release_gl_objects(self.queue, [self.cl_positions, self.cl_colors])

# <h2>Main class</h2>

# We wrap all the OpenGL interaction in a class. The class will contain an
# <code>__init__()</code> method to set up all OpenGL objects, any required
# callback methods, as well as a <code>run()</code> method to trigger execution
# of the GLUT main loop.
class OpenCLExample(object):
    # <h3>Initialization</h3>

    # When a <code>OpenCLExample</code> instance is created, we need to
    # initialize a few OpenGL objects.
    def __init__(self):
        # First, we create a window; this also creates an OpenGL context.
        self.window = GlutWindow(double=True, alpha=True, depth=True)

        # Then, we set the GLUT display and keyboard callback functions which
        # will be defined later.
        self.window.display_callback = self.display
        self.window.keyboard_callback = self.keyboard

        # Here, we generate numpy arrays to hold the positions, colors, and
        # velocities of the particles:
        num = 200000
        positions = numpy.empty((num, 4), dtype=numpy.float32)
        colors = numpy.empty((num, 4), dtype=numpy.float32)
        velocities = numpy.empty((num, 4), dtype=numpy.float32)

        # So far, the array contents are undefined. We have to initialize them with meaningful values:
        positions[:, 0] = numpy.sin(numpy.arange(0, num) * 2 * numpy.pi / num) * (numpy.random.random_sample((num,)) / 3 + 0.2)
        positions[:, 1] = numpy.cos(numpy.arange(0, num) * 2 * numpy.pi / num) * (numpy.random.random_sample((num,)) / 3 + 0.2)
        positions[:, 2:] = 0, 1
        colors[:] = 0, 1, 0, 1
        velocities[:, :2] = 2 * positions[:, :2]
        velocities[:, 2] = 3
        velocities[:, 3] = numpy.random.random_sample((num,))

        # Instead of simply generating a vertex array from the position and color
        # data, we first generate array buffers for them:
        gl_positions = ArrayBuffer(data=positions, usage="DYNAMIC_DRAW")
        gl_colors = ArrayBuffer(data=colors, usage="DYNAMIC_DRAW")
        # These array buffers will later also be used by OpenCL. We do not need to
        # wrap <code>velocities</code> in this way, as it will only be used by
        # OpenCL and can be wrapped in an OpenCL buffer directly.

        # We now create a vertex array that will pass the position and color data
        # to the shader. The vertex array constructor accepts
        # <code>ArrayBuffer</code> instances:
        self.vao = VertexArray(gl_positions, gl_colors)

        # In the OpenGL core profile, there is no such thing as a "standard pipeline"
        # any more. We use the minimalistic <code>defaultpipeline</code> from the
        # <code>glitter.convenience</code> module to create a shader program instead:
        self.shader = get_default_program()
        
        # Here, we create the <code>CLCode</code> object that manages OpenCL
        # interaction. It is passed the OpenGL buffer objects as well as a numpy
        # array of velocities.
        self.clcode = CLCode(gl_positions, gl_colors, velocities)

    # <h3>Callback functions</h3>

    # <h4>Display function</h4>

    # Here we define the display function. It will be called by GLUT whenever the
    # screen has to be redrawn.
    def display(self):
        # First we clear the default framebuffer:
        self.window.clear()
        
        # To draw the vertex array, we use:
        self.vao.draw()

        # After all rendering commands have been issued, we swap the back buffer to
        # the front, making the rendered image visible all at once:
        self.window.swap_buffers()

    # <h4>Timer function</h4>

    # The animation is controlled by a GLUT timer. The timer callback animates the
    # particle system, schedules the next timer event, and causes a screen redraw:
    def timer(self):
        # We first tell an instance of the <code>CLCode</code> class to execute the
        # OpenCL kernel:
        self.clcode.execute(10)

        # The following line schedules the next timer event to execute after one millisecond.
        self.window.add_timer(1, self.timer)

        # Finally, we tell GLUT to redraw the screen.
        self.window.post_redisplay()

    # <h4>Keyboard function</h4>

    # To further illustrate the concept of GLUT callbacks, here's a keyboard
    # handler that will simply make the program exit when any key is pressed:
    def keyboard(self, key, x, y):
        raise SystemExit

    # <h3>Running</h3>
    
    # We will call the <code>run()</code> method later to run the OpenGL code.
    def run(self):
        # To start the animation, we call the timer once; all subsequent timer
        # calls will be scheduled by the timer function itself.
        self.timer()

        # The default program is bound by using a <code>with</code> statement. At
        # the same time, we can pass in additional uniform variables, such as the
        # modelview matrix:
        with self.shader(modelview_matrix=((1, 0, 0, 0), (0, 0, 1, 0), (0, 1, 0, 0), (0, 0, 0, 2))):
            # With the shader bound, we enter the GLUT main loop.
            main_loop()

        # When the main loop exits, control is handed back to the script,
        # unless <code>SystemExit</code> has been raised by the keyboard
        # handler.

# <h2>Main section</h2>

# Finally, if this program is being run from the command line, we instanciate
# the main class and run it.
if __name__ == "__main__":
    OpenCLExample().run()




"""Error classes.

@author: Stephan Wenger
@date: 2012-02-29
"""

class GlitterError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message.decode("ascii")

class ShaderError(GlitterError):
    pass

class ShaderCompileError(ShaderError):
    pass

class ShaderLinkError(ShaderError):
    pass

class ShaderValidateError(ShaderError):
    pass




"""Descriptors for per-context state with complicated setters.

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils import blend_functions, blend_equations

class BlendFuncProxy(object):
    def __init__(self, arg):
        self._arg = arg

    def __get__(self, obj, cls=None):
        _value = _gl.GLint()
        with obj:
            _gl.glGetIntegerv(self._arg, _gl.pointer(_value))
        return blend_functions[_value.value]

    def __set__(self, obj, value):
        src_rgb = blend_functions(value) if self._arg == _gl.GL_BLEND_SRC_RGB else obj.blend_src_rgb
        dst_rgb = blend_functions(value) if self._arg == _gl.GL_BLEND_DST_RGB else obj.blend_dst_rgb
        src_alpha = blend_functions(value) if self._arg == _gl.GL_BLEND_SRC_ALPHA else obj.blend_src_alpha
        dst_alpha = blend_functions(value) if self._arg == _gl.GL_BLEND_DST_ALPHA else obj.blend_dst_alpha
        with obj:
            _gl.glBlendFuncSeparate(src_rgb._value, dst_rgb._value, src_alpha._value, dst_alpha._value)

    def __repr__(self):
        return "proxy for %s" % self._arg

class BlendEquationProxy(object):
    def __init__(self, arg):
        self._arg = arg

    def __get__(self, obj, cls=None):
        _value = _gl.GLint()
        with obj:
            _gl.glGetIntegerv(self._arg, _gl.pointer(_value))
        return blend_equations[_value.value]

    def __set__(self, obj, value):
        mode_rgb = blend_equations(value) if self._arg == _gl.GL_BLEND_EQUATION_RGB else obj.blend_equation_rgb
        mode_alpha = blend_equations(value) if self._arg == _gl.GL_BLEND_EQUATION_ALPHA else obj.blend_equation_alpha
        with obj:
            _gl.glBlendEquationSeparate(mode_rgb._value, mode_alpha._value)

    def __repr__(self):
        return "proxy for %s" % self._arg

class PolygonOffsetProxy(object):
    def __init__(self, arg):
        self._arg = arg

    def __get__(self, obj, cls=None):
        _value = _gl.GLfloat()
        with obj:
            _gl.glGetFloatv(self._arg, _gl.pointer(_value))
        return _value.value

    def __set__(self, obj, value):
        factor = value if self._arg == _gl.GL_POLYGON_OFFSET_FACTOR else obj.polygon_offset_factor
        units = value if self._arg == _gl.GL_POLYGON_OFFSET_UNITS else obj.polygon_offset_units
        with obj:
            _gl.glPolygonOffset(factor, units)

    def __repr__(self):
        return "proxy for %s" % self._arg




"""Framebuffer object class.

@warn: Framebuffers are currently bound for drawing only (C{GL_DRAW_FRAMEBUFFER}, not C{GL_READ_FRAMEBUFFER}).
@todo: Implement binding for reading; wrap C{glReadPixels}, C{glBlitFramebuffer}, C{glCopyTexSubImage} and related framebuffer copy functions.
@todo: Add properties for C{viewport}, C{color_writemask}, C{depth_writemask},
C{blend_func}, C{blend_equation}, and C{depth_range} (using
C{glDepthRangeArray} and C{glDepthRangeIndexed}).

@author: Stephan Wenger
@date: 2012-02-29
"""

import glitter.raw as _gl
from glitter.utils import ManagedObject, BindableObject, framebuffer_status

class Framebuffer(ManagedObject, BindableObject):
    _generate_id = _gl.glGenFramebuffers
    _delete_id = _gl.glDeleteBuffers
    _db = "framebuffers"
    _binding = "draw_framebuffer_binding"
    _target = _gl.GL_DRAW_FRAMEBUFFER

    framebuffer_status = framebuffer_status
    _initialized = False

    def __init__(self, *attachments, **kwargs):
        """Create a new framebuffer.

        @param attachments: Textures to bind to color attachments.
        @type attachments: C{list} of L{Texture}s
        @param kwargs: Named arguments.
        @type kwargs: C{dict}
        @keyword context: The context in which to create the framebuffer.
        @type context: L{Context}
        @keyword depth: An optional depth buffer attachment.
        @type depth: L{Texture}
        @keyword stencil: An optional stencil buffer attachment.
        @type stencil: L{Texture}
        """

        super(Framebuffer, self).__init__(context=kwargs.pop("context", None))
        self._attachments = {}
        
        if isinstance(attachments, dict):
            attachments = dict(attachments)
        else:
            attachments = dict(enumerate(attachments))
        for i in range(self._context.max_color_attachments):
            self[i] = attachments.pop(i, None)
        if attachments:
            raise ValueError("framebuffer has no attachment(s) %s" % ", ".join("'%s'" % x for x in list(attachments.keys())))

        self.depth = kwargs.pop("depth", None)
        self.stencil = kwargs.pop("stencil", None)

        if kwargs:
            raise TypeError("__init__() got an unexpected keyword argument '%s'" % list(kwargs.keys())[0])

        self._initialized = True

    def __getitem__(self, index):
        return self._attachments[index]

    def __setitem__(self, index, value):
        self.attach(index, value)

    def __delitem__(self, index):
        self.attach(index, None)

    def _on_bind(self):
        """Setup global framebuffer related state.

        @todo: Setup C{color_writemasks}, C{depth_writemasks}, C{blend_funcs}, C{blend_equations}, and C{depth_ranges} after C{viewport}.
        """

        if self._initialized:
            self._stack.append(self._context.draw_buffers)
            self._context.draw_buffers = [i if self[i] is not None else None for i in range(self._context.max_color_attachments)]

            self._stack.append(self._context.viewport)
            if self.shape is not None:
                self._context.viewport = (0, 0, self.shape[1], self.shape[0])

    def _on_release(self):
        """Restore global framebuffer related state.

        @todo: Restore C{depth_ranges}, C{blend_equations}, C{blend_funcs}, C{depth_writemasks}, and C{color_writemasks} before C{viewport}.
        """

        if self._initialized:
            self._context.viewport = self._stack.pop()
            self._context.draw_buffers = self._stack.pop()

    def _attach(self, attachment, texture=None, layer=None, level=0):
        """Attach a texture to an attachment.

        C{texture} may be a (texture, layer) tuple generated by
        L{Texture}C{.__getitem__()}.
        """

        if type(texture) is tuple:
            if layer is not None:
                raise ValueError("cannot provide layer as both keyword and texture tuple")
            texture, layer = texture
            return self._attach(attachment, texture, layer, level)

        with self:
            if texture is None:
                _gl.glFramebufferTextureLayer(self._target, attachment, 0, level, 0)
            elif layer is None:
                _gl.glFramebufferTexture(self._target, attachment, texture._id, level)
            else:
                _gl.glFramebufferTextureLayer(self._target, attachment, texture._id, level, layer)

    def attach(self, index, texture=None, layer=None, level=0):
        """Attach a texture to color attachment C{index}.

        C{texture} may be a (texture, layer) tuple generated by
        L{Texture}C{.__getitem__()}.  Otherwise, C{layer} specifies which
        texture layer is to be bound. All layers are bound when C{layer} is
        C{None}.

        C{level} specifies which resolution level of the texture is to be
        bound.

        If C{texture} is C{None}, the attachment will be unbound.
        """

        self._attach(_gl.GL_COLOR_ATTACHMENT0 + index, texture, layer, level)
        self._attachments[index] = texture

    @property
    def depth(self):
        if not hasattr(self, "_depth"):
            return None
        return self._depth

    @depth.setter
    def depth(self, depth):
        self.attach_depth(depth)

    @depth.deleter
    def depth(self):
        self.attach_depth(None)

    def attach_depth(self, texture=None, layer=None, level=0):
        """Attach a texture to the depth attachment.

        C{texture} may be a (texture, layer) tuple generated by
        L{Texture}C{.__getitem__()}.  Otherwise, C{layer} specifies which
        texture layer is to be bound. All layers are bound when C{layer} is
        C{None}.

        C{level} specifies which resolution level of the texture is to be
        bound.

        If C{texture} is C{None}, the attachment will be unbound.
        """

        self._attach(_gl.GL_DEPTH_ATTACHMENT, texture, layer, level)
        self._depth = texture

    @property
    def stencil(self):
        if not hasattr(self, "_stencil"):
            return None
        return self._stencil

    @stencil.setter
    def stencil(self, stencil):
        self.attach_stencil(stencil)

    @stencil.deleter
    def stencil(self):
        self.attach_stencil(None)

    def attach_stencil(self, texture=None, layer=None, level=0):
        """Attach a texture to the stencil attachment.

        C{texture} may be a (texture, layer) tuple generated by
        L{Texture}C{.__getitem__()}.  Otherwise, C{layer} specifies which
        texture layer is to be bound. All layers are bound when C{layer} is
        C{None}.

        C{level} specifies which resolution level of the texture is to be
        bound.

        If C{texture} is C{None}, the attachment will be unbound.
        """

        self._attach(_gl.GL_STENCIL_ATTACHMENT, texture, layer, level)
        self._stencil = texture

    @property
    def status(self):
        """The framebuffer completeness status.

        @rtype: L{Framebuffer.framebuffer_status} enum
        """

        with self:
            return self.framebuffer_status[_gl.glCheckFramebufferStatus(self._target)]

    @property
    def shape(self):
        shape = None
        for i, attachment in sorted(self._attachments.items()):
            if attachment is not None:
                if type(attachment) is tuple:
                    texture, layer = attachment
                    shape = (min(shape[0], texture.shape[1]), min(shape[1], texture.shape[2])) if shape else texture.shape[1:3]
                else:
                    shape = (min(shape[0], attachment.shape[0]), min(shape[1], attachment.shape[1])) if shape else attachment.shape[:2]
        if self.depth is not None:
            shape = (min(shape[0], self.depth.shape[0]), min(shape[1], self.depth.shape[1])) if shape else self.depth.shape[:2]
        if self.stencil is not None:
            shape = (min(shape[0], self.stencil.shape[0]), min(shape[1], self.stencil.shape[1])) if shape else self.stencil.shape[:2]
        return shape

    def clear(self, color=None, depth=None, stencil=None):
        """Clear the framebuffer.

        @param color: Whether to clear the color buffer, and optionally, to which value.
        @type color: C{bool} or C{numpy.ndarray}.
        @param depth: Whether to clear the depth buffer, and optionally, to which value.
        @type depth: C{bool} or C{numpy.ndarray}.
        @param stencil: Whether to clear the stencil buffer, and optionally, to which value.
        @type stencil: C{bool} or C{numpy.ndarray}.

        If no parameters are given, color, depth and stencil are cleared with
        the current clear values.

        @todo: Use C{glClearBuffer} to clear selected attachments only.
        """

        with self:
            self._context._perform_gl_clear(color, depth, stencil)

__all__ = ["Framebuffer"]




import numpy

from glitter import EnumConstant
from glitter.contexts import context_manager

def check_property(context, name):
    value = getattr(context, name)
    try:
        if isinstance(value, EnumConstant):
            if name in ("draw_buffer", "read_buffer"):
                return # avoid problems with unavailable stereo buffers
            valid_values = list(value._enum._reverse_dict.values())
            for value in valid_values:
                setattr(context, name, value)
                assert numpy.all(getattr(context, name) == value), "property %s is broken" % name
        else:
            if type(value) is float:
                value *= 0.5
                setattr(context, name, value)
                assert numpy.all(getattr(context, name) == value), "property %s is broken" % name
                value += 0.5
                setattr(context, name, value)
                assert numpy.all(getattr(context, name) == value), "property %s is broken" % name
    except AttributeError:
        pass # "AttributeError: can't set attribute" is okay for read-only attributes

def test_property_generator():
    context = context_manager.current_context or context_manager.create_default_context()
    properties = [x for x in dir(context) if not x.startswith("_")]

    for p in properties:
        yield check_property, context, p




"""Convenience factory methods for geometric transform matrices.

@author: Stephan Wenger
@date: 2012-03-05
"""

import numpy as _np

def identity_matrix():
    return _np.matrix(_np.eye(4))

def rotation_matrix(angle, axis, degrees=False):
    if degrees:
        angle = _np.deg2rad(angle)
    c, s = _np.cos(angle), _np.sin(angle)
    x, y, z = axis / _np.sqrt(sum(a ** 2 for a in axis))
    return _np.matrix((
        (x * x * (1 - c) + c,     x * y * (1 - c) - z * s, x * z * (1 - c) + y * s, 0),
        (y * x * (1 - c) + z * s, y * y * (1 - c) + c,     y * z * (1 - c) - x * s, 0),
        (z * x * (1 - c) - y * s, z * y * (1 - c) + x * s, z * z * (1 - c) + c,     0),
        (0,                       0,                       0,                       1),
        ))

def scale_matrix(*s):
    if len(s) == 3:
        x, y, z = s
    elif len(s) == 1:
        x = y = z = s[0]
    else:
        raise ValueError("invalid scaling parameter")
    return _np.matrix(((x, 0, 0, 0), (0, y, 0, 0), (0, 0, z, 0), (0, 0, 0, 1)))

__all__ = ["identity_matrix", "rotation_matrix", "scale_matrix"]




"""Shader program emulating the traditional OpenGL default pipeline.

@author: Stephan Wenger
@date: 2012-03-06
"""

from glitter.shaders import ShaderProgram

vertex_code = """
#version 400 core

#define MODELVIEW 1
#define COLOR 1

layout(location=0) in vec4 in_position;
#if COLOR
layout(location=1) in vec4 in_color;
#endif
#if MODELVIEW
uniform mat4 modelview_matrix;
#endif
out vec4 ex_color;

void main() {
    #if MODELVIEW
    gl_Position = modelview_matrix * in_position;
    #else
    gl_Position = in_position;
    #endif
    #if COLOR
    ex_color = in_color;
    #else
    ex_color = vec4(1.0);
    #endif
}
"""
"""Default vertex shader."""

fragment_code = """
#version 400 core

in vec4 ex_color;
layout(location=0) out vec4 out_color;

void main() {
    out_color = ex_color;
}
"""
"""Default fragment shader."""

def get_default_program(modelview=True, color=True, context=None):
    """Get a shader program emulating the default pipeline.

    The program has two inputs, C{in_position} and (optionally) C{in_color};
    one output, C{out_color}; and (optionally) one uniform variable,
    C{modelview_matrix}.

    @param modelview: Whether to use the modelview matrix.
    @type modelview: C{bool}
    @param color: Whether to use colors.
    @type color: C{bool}
    @param context: The context to create the program in, or C{None} for the current context.
    @type context: L{Context}
    @rtype: L{ShaderProgram}
    """

    vertex = vertex_code
    if not modelview:
        vertex = vertex.replace("#define MODELVIEW 1", "#define MODELVIEW 0")
    if not color:
        vertex = vertex.replace("#define COLOR 1", "#define COLOR 0")
    return ShaderProgram(vertex=vertex, fragment=fragment_code, context=context)

__all__ = ["vertex_code", "fragment_code", "get_default_program"]




"""Element array buffer class.

@author: Stephan Wenger
@date: 2012-02-29
"""

import numpy as _np

import glitter.raw as _gl
from glitter.utils import buffer_dimensions_to_primitive, Datatype, coerce_array
from glitter.arrays.basebuffer import BaseBuffer

class ElementArrayBuffer(BaseBuffer):
    _binding = "element_array_buffer_binding"
    _target = _gl.GL_ELEMENT_ARRAY_BUFFER

    def set_data(self, data=None, shape=None, dtype=None, usage=None):
        if data is not None:
            data = coerce_array(data, dtype, force_integer=True, force_unsigned=True)
            dtype = Datatype.from_numpy(data.dtype)
        if dtype.is_signed() or not dtype.is_integer():
            raise TypeError("%s must be of unsigned integer type" % self.__class__.__name__)
        super(ElementArrayBuffer, self).set_data(data, shape, dtype, usage)

    def draw(self, mode=None, count=None, first=0, instances=None):
        if mode is None:
            if len(self.shape) >= 2:
                mode = buffer_dimensions_to_primitive.get(self.shape[-1], None)
        if mode is None:
            raise ValueError("must specify mode")
        if count is None:
            count = _np.prod(self.shape)
        if instances is None:
            with self:
                _gl.glDrawElements(mode._value, count, self.dtype._as_gl(), first * self.dtype.nbytes)
        else:
            with self:
                _gl.glDrawElementsInstanced(mode._value, count, self.dtype._as_gl(), first * self.dtype.nbytes, instances)

__all__ = ["ElementArrayBuffer"]




"""Descriptors for texture image units.

@author: Stephan Wenger
@date: 2012-02-29
"""

import itertools as _itertools

import glitter.raw as _gl
from glitter.utils import BindableObject, GlitterError
from glitter.contexts.proxies import BindingProxy

try:
    next
except NameError:
    def next(iterator):
        return iterator.next()

class TextureUnit(BindableObject):
    """Proxy for a texture unit.
    
    @warning: For internal use by L{TextureUnitList} only. Do not instantiate
    or use directly; use the methods of L{TextureUnitList} instead.
    """

    _binding = "active_texture"

    def __init__(self, context, _id):
        super(TextureUnit, self).__init__(context=context)
        self._context = context
        self._id = _id
        self._use_count = 0

    def __str__(self):
        return "GL_TEXTURE%d" % (self._id - _gl.GL_TEXTURE0)

    def activate(self):
        self._context.active_texture = self

    def is_available(self):
        return self._use_count == 0

    texture_binding_1d                   = BindingProxy(_gl.glBindTexture,         [_gl.GL_TEXTURE_1D                  ])
    texture_binding_1d_array             = BindingProxy(_gl.glBindTexture,         [_gl.GL_TEXTURE_1D_ARRAY            ])
    texture_binding_2d                   = BindingProxy(_gl.glBindTexture,         [_gl.GL_TEXTURE_2D                  ])
    texture_binding_2d_array             = BindingProxy(_gl.glBindTexture,         [_gl.GL_TEXTURE_2D_ARRAY            ])
    texture_binding_2d_multisample       = BindingProxy(_gl.glBindTexture,         [_gl.GL_TEXTURE_2D_MULTISAMPLE      ])
    texture_binding_2d_multisample_array = BindingProxy(_gl.glBindTexture,         [_gl.GL_TEXTURE_2D_MULTISAMPLE_ARRAY])
    texture_binding_3d                   = BindingProxy(_gl.glBindTexture,         [_gl.GL_TEXTURE_3D                  ])
    texture_binding_buffer               = BindingProxy(_gl.glBindTexture,         [_gl.GL_TEXTURE_BUFFER              ])
    texture_binding_cube_map             = BindingProxy(_gl.glBindTexture,         [_gl.GL_TEXTURE_CUBE_MAP            ])
    texture_binding_rectangle            = BindingProxy(_gl.glBindTexture,         [_gl.GL_TEXTURE_RECTANGLE           ])
    sampler_binding                      = BindingProxy(_gl.glBindSampler,         ["_unit"                            ])

class TextureUnitList(object):
    """Proxy for the list of available texture units.
    
    @warning: C{TextureUnitList}s are created by their respective context only.
    Do not instantiate them directly. Using the C{TextureUnitList}'s L{bind}
    and L{release} methods directly is usually not necessary; L{Shader}s bind
    their textures automatically, and the L{bind<Texture.bind>} method of
    L{Texture}s is delegated to the C{TextureUnitList}.
    """

    def __init__(self, context):
        self._context = context
        num_units = context.max_combined_texture_image_units
        if not 0 < num_units < 4096: # sanity check
            raise GlitterError("implausible number of texture units detected; are you sure there is a current OpenGL context?")
        self._texture_units = [TextureUnit(context, _gl.GL_TEXTURE0 + i) for i in range(num_units)]
        self._context.active_texture = self[0]
        self._bound_textures = dict()

    def __getitem__(self, index):
        return self._texture_units[index]

    def __len__(self):
        return len(self._texture_units)

    def __str__(self):
        return "[%d texture units]" % len(self)

    def __repr__(self):
        return str(self)

    def bind(self, texture):
        """Bind `texture` to a free unit and return the unit id."""
        if texture in self._bound_textures:
            unit = self._bound_textures[texture]
            unit._use_count += 1
        else:
            try:
                unit = next(_itertools.dropwhile(lambda x: not x.is_available(), self))
            except StopIteration:
                raise RuntimeError("no free texture units available")
            setattr(unit, texture._binding, texture)
            self._bound_textures[texture] = unit
            unit._use_count = 1
        return unit._id - _gl.GL_TEXTURE0

    def release(self, texture):
        """Unbind `texture`."""
        unit = self._bound_textures[texture]
        if unit._use_count  == 1:
            setattr(unit, texture._binding, None)
            del self._bound_textures[texture]
            unit._use_count = 0
        else:
            unit._use_count -= 1

